# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A set of standard astronomical equivalencies.

The equivalency class and all equivalency functions defined here are also
available in (and should be used through) the `astropy.units` namespace.

"""

import functools
from typing import Final

# THIRD-PARTY
import numpy as np

# LOCAL
from astropy.constants import si as _si
from astropy.utils import deprecated_renamed_argument

from . import astrophys, cgs, dimensionless_unscaled, misc, si
from .core import Unit
from .errors import UnitsError
from .function import units as function_units
from .photometric import maggy

__all__ = [
    "Equivalency",
    "beam_angular_area",
    "brightness_temperature",
    "dimensionless_angles",
    "doppler_optical",
    "doppler_radio",
    "doppler_redshift",
    "doppler_relativistic",
    "logarithmic",
    "magnetic_flux_field",
    "mass_energy",
    "molar_mass_amu",
    "parallax",
    "pixel_scale",
    "plate_scale",
    "spectral",
    "spectral_density",
    "temperature",
    "temperature_energy",
    "thermodynamic_temperature",
    "zero_point_flux",
]

km_per_s: Final = si.km / si.s
ckms: Final = _si.c.to_value(km_per_s)


class Equivalency(list):
    """
    A container for a units equivalency.

    Attributes
    ----------
    name: `str`
        The name of the equivalency.
    kwargs: `dict`
        Any positional or keyword arguments used to make the equivalency.
    """

    def __init__(self, equiv_list, name="", kwargs=None):
        super().__init__(equiv_list)
        self.name = [name]
        self.kwargs = [kwargs] if kwargs is not None else [{}]

    def __add__(self, other):
        if isinstance(other, Equivalency):
            # The super() returns a list, which is really a bit weird,
            # but that means we have to pass it back through the initializer.
            new = self.__class__(super().__add__(other))
            # Avoid the change to list of the name and kwargs arguments.
            new.name = self.name + other.name
            new.kwargs = self.kwargs + other.kwargs
            return new
        else:
            return super().__add__(other)  # Let list take care.

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.name == other.name
            and self.kwargs == other.kwargs
        )


@functools.cache
def dimensionless_angles():
    """Allow angles to be equivalent to dimensionless (with 1 rad = 1 m/m = 1).

    It is special compared to other equivalency pairs in that it
    allows this independent of the power to which the angle is raised,
    and independent of whether it is part of a more complicated unit.
    """
    return Equivalency([(si.radian, None)], "dimensionless_angles")


@functools.cache
def logarithmic():
    """Allow logarithmic units to be converted to dimensionless fractions."""
    return Equivalency(
        [(dimensionless_unscaled, function_units.dex, np.log10, lambda x: 10.0**x)],
        "logarithmic",
    )


@functools.cache
def parallax():
    """
    Returns a list of equivalence pairs that handle the conversion
    between parallax angle and distance.
    """

    def parallax_converter(x):
        x = np.asanyarray(x)
        d = 1 / x

        if np.iterable(d):
            d[d < 0] = np.nan
            return d

        else:
            if d < 0:
                return np.array(np.nan)
            else:
                return d

    return Equivalency(
        [(si.arcsecond, astrophys.parsec, parallax_converter)], "parallax"
    )


@functools.cache
def spectral():
    """
    Returns a list of equivalence pairs that handle spectral
    wavelength, wave number, frequency, and energy equivalencies.

    Allows conversions between wavelength units, wave number units,
    frequency units, and energy units as they relate to light.

    There are two types of wave number:

        * spectroscopic - :math:`1 / \\lambda` (per meter)
        * angular - :math:`2 \\pi / \\lambda` (radian per meter)

    """
    c = _si.c.value
    h = _si.h.value
    hc = h * c
    two_pi = 2.0 * np.pi
    inv_m_spec = si.m**-1
    inv_m_ang = si.radian / si.m

    return Equivalency(
        [
            (si.m, si.Hz, lambda x: c / x),
            (si.m, si.J, lambda x: hc / x),
            (si.Hz, si.J, lambda x: h * x, lambda x: x / h),
            (si.m, inv_m_spec, lambda x: 1.0 / x),
            (si.Hz, inv_m_spec, lambda x: x / c, lambda x: c * x),
            (si.J, inv_m_spec, lambda x: x / hc, lambda x: hc * x),
            (inv_m_spec, inv_m_ang, lambda x: x * two_pi, lambda x: x / two_pi),
            (si.m, inv_m_ang, lambda x: two_pi / x),
            (si.Hz, inv_m_ang, lambda x: two_pi * x / c, lambda x: c * x / two_pi),
            (si.J, inv_m_ang, lambda x: x * two_pi / hc, lambda x: hc * x / two_pi),
        ],
        "spectral",
    )


@deprecated_renamed_argument(
    "factor", None, since="7.0", alternative='"wav" as a "Quantity"'
)
def spectral_density(wav, factor=None):
    """
    Returns a list of equivalence pairs that handle spectral density
    with regard to wavelength and frequency.

    Parameters
    ----------
    wav : `~astropy.units.Quantity`
        `~astropy.units.Quantity` associated with values being converted
        (e.g., wavelength or frequency).
    factor : array_like
        If ``wav`` is a |Unit| instead of a |Quantity| then ``factor``
        is the value ``wav`` will be multiplied with to convert it to
        a |Quantity|.

        .. deprecated:: 7.0

            ``factor`` is deprecated. Pass in ``wav`` as a |Quantity|,
            not as a |Unit|.
    """
    from .core import UnitBase

    if isinstance(wav, UnitBase):
        if factor is None:
            raise ValueError("If `wav` is specified as a unit, `factor` should be set")
        wav = factor * wav  # Convert to Quantity
    c_Aps = _si.c.to_value(si.AA / si.s)  # Angstrom/s
    h_cgs = _si.h.cgs.value  # erg * s
    hc = c_Aps * h_cgs

    # flux density
    f_la = cgs.erg / si.angstrom / si.cm**2 / si.s
    f_nu = cgs.erg / si.Hz / si.cm**2 / si.s
    nu_f_nu = cgs.erg / si.cm**2 / si.s
    la_f_la = nu_f_nu
    phot_f_la = astrophys.photon / (si.cm**2 * si.s * si.AA)
    phot_f_nu = astrophys.photon / (si.cm**2 * si.s * si.Hz)
    la_phot_f_la = astrophys.photon / (si.cm**2 * si.s)

    # luminosity density
    L_nu = cgs.erg / si.s / si.Hz
    L_la = cgs.erg / si.s / si.angstrom
    nu_L_nu = cgs.erg / si.s
    la_L_la = nu_L_nu
    phot_L_la = astrophys.photon / (si.s * si.AA)
    phot_L_nu = astrophys.photon / (si.s * si.Hz)

    # surface brightness (flux equiv)
    S_la = cgs.erg / si.angstrom / si.cm**2 / si.s / si.sr
    S_nu = cgs.erg / si.Hz / si.cm**2 / si.s / si.sr
    nu_S_nu = cgs.erg / si.cm**2 / si.s / si.sr
    la_S_la = nu_S_nu
    phot_S_la = astrophys.photon / (si.cm**2 * si.s * si.AA * si.sr)
    phot_S_nu = astrophys.photon / (si.cm**2 * si.s * si.Hz * si.sr)

    # surface brightness (luminosity equiv)
    SL_nu = cgs.erg / si.s / si.Hz / si.sr
    SL_la = cgs.erg / si.s / si.angstrom / si.sr
    nu_SL_nu = cgs.erg / si.s / si.sr
    la_SL_la = nu_SL_nu
    phot_SL_la = astrophys.photon / (si.s * si.AA * si.sr)
    phot_SL_nu = astrophys.photon / (si.s * si.Hz * si.sr)

    def f_la_to_f_nu(x):
        return x * (wav.to_value(si.AA, spectral()) ** 2 / c_Aps)

    def f_la_from_f_nu(x):
        return x / (wav.to_value(si.AA, spectral()) ** 2 / c_Aps)

    def f_nu_to_nu_f_nu(x):
        return x * wav.to_value(si.Hz, spectral())

    def f_nu_from_nu_f_nu(x):
        return x / wav.to_value(si.Hz, spectral())

    def f_la_to_la_f_la(x):
        return x * wav.to_value(si.AA, spectral())

    def f_la_from_la_f_la(x):
        return x / wav.to_value(si.AA, spectral())

    def phot_f_la_to_f_la(x):
        return hc * x / wav.to_value(si.AA, spectral())

    def phot_f_la_from_f_la(x):
        return x * wav.to_value(si.AA, spectral()) / hc

    def phot_f_la_to_f_nu(x):
        return h_cgs * x * wav.to_value(si.AA, spectral())

    def phot_f_la_from_f_nu(x):
        return x / (wav.to_value(si.AA, spectral()) * h_cgs)

    def phot_f_la_to_phot_f_nu(x):
        return x * wav.to_value(si.AA, spectral()) ** 2 / c_Aps

    def phot_f_la_from_phot_f_nu(x):
        return c_Aps * x / wav.to_value(si.AA, spectral()) ** 2

    phot_f_nu_to_f_nu = phot_f_la_to_f_la
    phot_f_nu_from_f_nu = phot_f_la_from_f_la

    def phot_f_nu_to_f_la(x):
        return x * hc * c_Aps / wav.to_value(si.AA, spectral()) ** 3

    def phot_f_nu_from_f_la(x):
        return x * wav.to_value(si.AA, spectral()) ** 3 / (hc * c_Aps)

    # for luminosity density
    L_nu_to_nu_L_nu = f_nu_to_nu_f_nu
    L_nu_from_nu_L_nu = f_nu_from_nu_f_nu
    L_la_to_la_L_la = f_la_to_la_f_la
    L_la_from_la_L_la = f_la_from_la_f_la

    phot_L_la_to_L_la = phot_f_la_to_f_la
    phot_L_la_from_L_la = phot_f_la_from_f_la
    phot_L_la_to_L_nu = phot_f_la_to_f_nu
    phot_L_la_from_L_nu = phot_f_la_from_f_nu
    phot_L_la_to_phot_L_nu = phot_f_la_to_phot_f_nu
    phot_L_la_from_phot_L_nu = phot_f_la_from_phot_f_nu
    phot_L_nu_to_L_nu = phot_f_nu_to_f_nu
    phot_L_nu_from_L_nu = phot_f_nu_from_f_nu
    phot_L_nu_to_L_la = phot_f_nu_to_f_la
    phot_L_nu_from_L_la = phot_f_nu_from_f_la

    return Equivalency(
        [
            # flux
            (f_la, f_nu, f_la_to_f_nu, f_la_from_f_nu),
            (f_nu, nu_f_nu, f_nu_to_nu_f_nu, f_nu_from_nu_f_nu),
            (f_la, la_f_la, f_la_to_la_f_la, f_la_from_la_f_la),
            (phot_f_la, f_la, phot_f_la_to_f_la, phot_f_la_from_f_la),
            (phot_f_la, f_nu, phot_f_la_to_f_nu, phot_f_la_from_f_nu),
            (phot_f_la, phot_f_nu, phot_f_la_to_phot_f_nu, phot_f_la_from_phot_f_nu),
            (phot_f_nu, f_nu, phot_f_nu_to_f_nu, phot_f_nu_from_f_nu),
            (phot_f_nu, f_la, phot_f_nu_to_f_la, phot_f_nu_from_f_la),
            # integrated flux
            (la_phot_f_la, la_f_la, phot_f_la_to_f_la, phot_f_la_from_f_la),
            # luminosity
            (L_la, L_nu, f_la_to_f_nu, f_la_from_f_nu),
            (L_nu, nu_L_nu, L_nu_to_nu_L_nu, L_nu_from_nu_L_nu),
            (L_la, la_L_la, L_la_to_la_L_la, L_la_from_la_L_la),
            (phot_L_la, L_la, phot_L_la_to_L_la, phot_L_la_from_L_la),
            (phot_L_la, L_nu, phot_L_la_to_L_nu, phot_L_la_from_L_nu),
            (phot_L_la, phot_L_nu, phot_L_la_to_phot_L_nu, phot_L_la_from_phot_L_nu),
            (phot_L_nu, L_nu, phot_L_nu_to_L_nu, phot_L_nu_from_L_nu),
            (phot_L_nu, L_la, phot_L_nu_to_L_la, phot_L_nu_from_L_la),
            # surface brightness (flux equiv)
            (S_la, S_nu, f_la_to_f_nu, f_la_from_f_nu),
            (S_nu, nu_S_nu, f_nu_to_nu_f_nu, f_nu_from_nu_f_nu),
            (S_la, la_S_la, f_la_to_la_f_la, f_la_from_la_f_la),
            (phot_S_la, S_la, phot_f_la_to_f_la, phot_f_la_from_f_la),
            (phot_S_la, S_nu, phot_f_la_to_f_nu, phot_f_la_from_f_nu),
            (phot_S_la, phot_S_nu, phot_f_la_to_phot_f_nu, phot_f_la_from_phot_f_nu),
            (phot_S_nu, S_nu, phot_f_nu_to_f_nu, phot_f_nu_from_f_nu),
            (phot_S_nu, S_la, phot_f_nu_to_f_la, phot_f_nu_from_f_la),
            # surface brightness (luminosity equiv)
            (SL_la, SL_nu, f_la_to_f_nu, f_la_from_f_nu),
            (SL_nu, nu_SL_nu, L_nu_to_nu_L_nu, L_nu_from_nu_L_nu),
            (SL_la, la_SL_la, L_la_to_la_L_la, L_la_from_la_L_la),
            (phot_SL_la, SL_la, phot_L_la_to_L_la, phot_L_la_from_L_la),
            (phot_SL_la, SL_nu, phot_L_la_to_L_nu, phot_L_la_from_L_nu),
            (phot_SL_la, phot_SL_nu, phot_L_la_to_phot_L_nu, phot_L_la_from_phot_L_nu),
            (phot_SL_nu, SL_nu, phot_L_nu_to_L_nu, phot_L_nu_from_L_nu),
            (phot_SL_nu, SL_la, phot_L_nu_to_L_la, phot_L_nu_from_L_la),
        ],
        "spectral_density",
        {"wav": wav, "factor": factor},
    )


def doppler_radio(rest):
    r"""
    Return the equivalency pairs for the radio convention for velocity.

    The radio convention for the relation between velocity and frequency is:

    :math:`V = c \frac{f_0 - f}{f_0}  ;  f(V) = f_0 ( 1 - V/c )`

    Parameters
    ----------
    rest : `~astropy.units.Quantity`
        Any quantity supported by the standard spectral equivalencies
        (wavelength, energy, frequency, wave number).

    References
    ----------
    `NRAO site defining the conventions <https://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html>`_

    Examples
    --------
    >>> import astropy.units as u
    >>> CO_restfreq = 115.27120*u.GHz  # rest frequency of 12 CO 1-0 in GHz
    >>> radio_CO_equiv = u.doppler_radio(CO_restfreq)
    >>> measured_freq = 115.2832*u.GHz
    >>> radio_velocity = measured_freq.to(u.km/u.s, equivalencies=radio_CO_equiv)
    >>> radio_velocity  # doctest: +FLOAT_CMP
    <Quantity -31.209092088877583 km / s>
    """
    assert_is_spectral_unit(rest)
    rest_in = functools.partial(rest.to_value, equivalencies=spectral())
    to_funcs = {
        misc.eV: lambda x: (1 - x / rest_in(misc.eV)) * ckms,
        si.Hz: lambda x: (1 - x / rest_in(si.Hz)) * ckms,
        si.AA: lambda x: (1 - rest_in(si.AA) / x) * ckms,
    }
    from_funcs = {
        misc.eV: lambda x: rest_in(misc.eV) * (1 - x / ckms),
        si.Hz: lambda x: rest_in(si.Hz) * (1 - x / ckms),
        si.AA: lambda x: rest_in(si.AA) / (1 - x / ckms),
    }
    return Equivalency(
        [
            (unit, km_per_s, to_func, from_funcs[unit])
            for unit, to_func in to_funcs.items()
        ],
        "doppler_radio",
        {"rest": rest},
    )


def doppler_optical(rest):
    r"""
    Return the equivalency pairs for the optical convention for velocity.

    The optical convention for the relation between velocity and frequency is:

    :math:`V = c \frac{f_0 - f}{f  }  ;  f(V) = f_0 ( 1 + V/c )^{-1}`

    Parameters
    ----------
    rest : `~astropy.units.Quantity`
        Any quantity supported by the standard spectral equivalencies
        (wavelength, energy, frequency, wave number).

    References
    ----------
    `NRAO site defining the conventions <https://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html>`_

    Examples
    --------
    >>> import astropy.units as u
    >>> CO_restfreq = 115.27120*u.GHz  # rest frequency of 12 CO 1-0 in GHz
    >>> optical_CO_equiv = u.doppler_optical(CO_restfreq)
    >>> measured_freq = 115.2832*u.GHz
    >>> optical_velocity = measured_freq.to(u.km/u.s, equivalencies=optical_CO_equiv)
    >>> optical_velocity  # doctest: +FLOAT_CMP
    <Quantity -31.20584348799674 km / s>
    """
    assert_is_spectral_unit(rest)
    rest_in = functools.partial(rest.to_value, equivalencies=spectral())
    to_funcs = {
        misc.eV: lambda x: (rest_in(misc.eV) / x - 1) * ckms,
        si.Hz: lambda x: (rest_in(si.Hz) / x - 1) * ckms,
        si.AA: lambda x: (x / rest_in(si.AA) - 1) * ckms,
    }
    from_funcs = {
        misc.eV: lambda x: rest_in(misc.eV) / (1 + x / ckms),
        si.Hz: lambda x: rest_in(si.Hz) / (1 + x / ckms),
        si.AA: lambda x: rest_in(si.AA) * (1 + x / ckms),
    }
    return Equivalency(
        [
            (unit, km_per_s, to_func, from_funcs[unit])
            for unit, to_func in to_funcs.items()
        ],
        "doppler_optical",
        {"rest": rest},
    )


def doppler_relativistic(rest):
    r"""
    Return the equivalency pairs for the relativistic convention for velocity.

    The full relativistic convention for the relation between velocity and frequency is:

    :math:`V = c \frac{f_0^2 - f^2}{f_0^2 + f^2} ;  f(V) = f_0 \frac{\left(1 - (V/c)^2\right)^{1/2}}{(1+V/c)}`

    Parameters
    ----------
    rest : `~astropy.units.Quantity`
        Any quantity supported by the standard spectral equivalencies
        (wavelength, energy, frequency, wave number).

    References
    ----------
    `NRAO site defining the conventions <https://www.gb.nrao.edu/~fghigo/gbtdoc/doppler.html>`_

    Examples
    --------
    >>> import astropy.units as u
    >>> CO_restfreq = 115.27120*u.GHz  # rest frequency of 12 CO 1-0 in GHz
    >>> relativistic_CO_equiv = u.doppler_relativistic(CO_restfreq)
    >>> measured_freq = 115.2832*u.GHz
    >>> relativistic_velocity = measured_freq.to(u.km/u.s, equivalencies=relativistic_CO_equiv)
    >>> relativistic_velocity  # doctest: +FLOAT_CMP
    <Quantity -31.207467619351537 km / s>
    >>> measured_velocity = 1250 * u.km/u.s
    >>> relativistic_frequency = measured_velocity.to(u.GHz, equivalencies=relativistic_CO_equiv)
    >>> relativistic_frequency  # doctest: +FLOAT_CMP
    <Quantity 114.79156866993588 GHz>
    >>> relativistic_wavelength = measured_velocity.to(u.mm, equivalencies=relativistic_CO_equiv)
    >>> relativistic_wavelength  # doctest: +FLOAT_CMP
    <Quantity 2.6116243681798923 mm>
    """
    assert_is_spectral_unit(rest)
    rest_in = functools.partial(rest.to_value, equivalencies=spectral())

    def to_vel_freq(x):
        restfreq2 = rest_in(si.Hz) ** 2
        return (restfreq2 - x**2) / (restfreq2 + x**2) * ckms

    def from_vel_freq(x):
        voverc = x / ckms
        return rest_in(si.Hz) * ((1 - voverc) / (1 + voverc)) ** 0.5

    def to_vel_wav(x):
        restwav2 = rest_in(si.AA) ** 2
        return (x**2 - restwav2) / (restwav2 + x**2) * ckms

    def from_vel_wav(x):
        voverc = x / ckms
        return rest_in(si.AA) * ((1 + voverc) / (1 - voverc)) ** 0.5

    def to_vel_en(x):
        resten2 = rest_in(misc.eV) ** 2
        return (resten2 - x**2) / (resten2 + x**2) * ckms

    def from_vel_en(x):
        voverc = x / ckms
        return rest_in(misc.eV) * ((1 - voverc) / (1 + voverc)) ** 0.5

    return Equivalency(
        [
            (si.Hz, km_per_s, to_vel_freq, from_vel_freq),
            (si.AA, km_per_s, to_vel_wav, from_vel_wav),
            (misc.eV, km_per_s, to_vel_en, from_vel_en),
        ],
        "doppler_relativistic",
        {"rest": rest},
    )


@functools.cache
def doppler_redshift():
    """
    Returns the equivalence between Doppler redshift (unitless) and radial velocity.

    .. note::

        This equivalency is not compatible with cosmological
        redshift in `astropy.cosmology.units`.

    """
    rv_unit = si.km / si.s
    C_KMS = _si.c.to_value(rv_unit)

    def convert_z_to_rv(z):
        zponesq = (1 + z) ** 2
        return C_KMS * (zponesq - 1) / (zponesq + 1)

    def convert_rv_to_z(rv):
        beta = rv / C_KMS
        return np.sqrt((1 + beta) / (1 - beta)) - 1

    return Equivalency(
        [(dimensionless_unscaled, rv_unit, convert_z_to_rv, convert_rv_to_z)],
        "doppler_redshift",
    )


@functools.cache
def molar_mass_amu():
    """
    Returns the equivalence between amu and molar mass.
    """
    return Equivalency([(si.g / si.mol, misc.u)], "molar_mass_amu")


@functools.cache
def mass_energy():
    """
    Returns a list of equivalence pairs that handle the conversion
    between mass and energy.
    """
    c2 = _si.c.value**2
    return Equivalency(
        [
            (si.kg, si.J, lambda x: x * c2, lambda x: x / c2),
            (si.kg / si.m**2, si.J / si.m**2, lambda x: x * c2, lambda x: x / c2),
            (si.kg / si.m**3, si.J / si.m**3, lambda x: x * c2, lambda x: x / c2),
            (si.kg / si.s, si.J / si.s, lambda x: x * c2, lambda x: x / c2),
        ],
        "mass_energy",
    )


def brightness_temperature(frequency, beam_area=None):
    r"""
    Defines the conversion between Jy/sr and "brightness temperature",
    :math:`T_B`, in Kelvins.  The brightness temperature is a unit very
    commonly used in radio astronomy.  See, e.g., "Tools of Radio Astronomy"
    (Wilson 2009) eqn 8.16 and eqn 8.19 (these pages are available on `google
    books
    <https://books.google.com/books?id=9KHw6R8rQEMC&pg=PA179&source=gbs_toc_r&cad=4#v=onepage&q&f=false>`__).

    :math:`T_B \equiv S_\nu / \left(2 k \nu^2 / c^2 \right)`

    If the input is in Jy/beam or Jy (assuming it came from a single beam), the
    beam area is essential for this computation: the brightness temperature is
    inversely proportional to the beam area.

    Parameters
    ----------
    frequency : `~astropy.units.Quantity`
        The observed ``spectral`` equivalent `~astropy.units.Unit` (e.g.,
        frequency or wavelength).  The variable is named 'frequency' because it
        is more commonly used in radio astronomy.
        BACKWARD COMPATIBILITY NOTE: previous versions of the brightness
        temperature equivalency used the keyword ``disp``, which is no longer
        supported.
    beam_area : `~astropy.units.Quantity` ['solid angle']
        Beam area in angular units, i.e. steradian equivalent

    Examples
    --------
    Arecibo C-band beam::

        >>> import numpy as np
        >>> from astropy import units as u
        >>> beam_sigma = 50*u.arcsec
        >>> beam_area = 2*np.pi*(beam_sigma)**2
        >>> freq = 5*u.GHz
        >>> equiv = u.brightness_temperature(freq)
        >>> (1*u.Jy/beam_area).to(u.K, equivalencies=equiv)  # doctest: +FLOAT_CMP
        <Quantity 3.526295144567176 K>

    VLA synthetic beam::

        >>> bmaj = 15*u.arcsec
        >>> bmin = 15*u.arcsec
        >>> fwhm_to_sigma = 1./(8*np.log(2))**0.5
        >>> beam_area = 2.*np.pi*(bmaj*bmin*fwhm_to_sigma**2)
        >>> freq = 5*u.GHz
        >>> equiv = u.brightness_temperature(freq)
        >>> (u.Jy/beam_area).to(u.K, equivalencies=equiv)  # doctest: +FLOAT_CMP
        <Quantity 217.2658703625732 K>

    Any generic surface brightness:

        >>> surf_brightness = 1e6*u.MJy/u.sr
        >>> surf_brightness.to(u.K, equivalencies=u.brightness_temperature(500*u.GHz)) # doctest: +FLOAT_CMP
        <Quantity 130.1931904778803 K>
    """
    nu = frequency.to(si.GHz, spectral())
    factor_Jy = (2 * _si.k_B * si.K * nu**2 / _si.c**2).to(astrophys.Jy).value
    factor_K = (astrophys.Jy / (2 * _si.k_B * nu**2 / _si.c**2)).to(si.K).value

    if beam_area is not None:
        beam = beam_area.to_value(si.sr)

        def convert_Jy_to_K(x_jybm):
            return x_jybm / beam / factor_Jy

        def convert_K_to_Jy(x_K):
            return x_K * beam / factor_K

        return Equivalency(
            [
                (astrophys.Jy, si.K, convert_Jy_to_K, convert_K_to_Jy),
                (astrophys.Jy / astrophys.beam, si.K, convert_Jy_to_K, convert_K_to_Jy),
            ],
            "brightness_temperature",
            {"frequency": frequency, "beam_area": beam_area},
        )
    else:

        def convert_JySr_to_K(x_jysr):
            return x_jysr / factor_Jy

        def convert_K_to_JySr(x_K):
            return x_K / factor_K  # multiplied by 1x for 1 steradian

        return Equivalency(
            [(astrophys.Jy / si.sr, si.K, convert_JySr_to_K, convert_K_to_JySr)],
            "brightness_temperature",
            {"frequency": frequency, "beam_area": beam_area},
        )


def beam_angular_area(beam_area):
    """
    Convert between the ``beam`` unit, which is commonly used to express the area
    of a radio telescope resolution element, and an area on the sky.
    This equivalency also supports direct conversion between ``Jy/beam`` and
    ``Jy/steradian`` units, since that is a common operation.

    Parameters
    ----------
    beam_area : unit-like
        The area of the beam in angular area units (e.g., steradians)
        Must have angular area equivalent units.
    """
    return Equivalency(
        [
            (astrophys.beam, Unit(beam_area)),
            (astrophys.beam**-1, Unit(beam_area) ** -1),
            (astrophys.Jy / astrophys.beam, astrophys.Jy / Unit(beam_area)),
        ],
        "beam_angular_area",
        {"beam_area": beam_area},
    )


def thermodynamic_temperature(frequency, T_cmb=None):
    r"""Defines the conversion between Jy/sr and "thermodynamic temperature",
    :math:`T_{CMB}`, in Kelvins.  The thermodynamic temperature is a unit very
    commonly used in cosmology. See eqn 8 in [1].

    :math:`K_{CMB} \equiv I_\nu / \left(2 k \nu^2 / c^2  f(\nu) \right)`

    with :math:`f(\nu) = \frac{ x^2 e^x}{(e^x - 1 )^2}`
    where :math:`x = h \nu / k T`

    Parameters
    ----------
    frequency : `~astropy.units.Quantity`
        The observed `spectral` equivalent `~astropy.units.Unit` (e.g.,
        frequency or wavelength). Must have spectral units.
    T_cmb :  `~astropy.units.Quantity` ['temperature'] or None
        The CMB temperature at z=0.  If `None`, the default cosmology will be
        used to get this temperature. Must have units of temperature.

    Notes
    -----
    For broad band receivers, this conversion do not hold
    as it highly depends on the frequency

    References
    ----------
    .. [1] Planck 2013 results. IX. HFI spectral response
       https://arxiv.org/abs/1303.5070

    Examples
    --------
    Planck HFI 143 GHz::

        >>> from astropy import units as u
        >>> from astropy.cosmology import Planck15
        >>> freq = 143 * u.GHz
        >>> equiv = u.thermodynamic_temperature(freq, Planck15.Tcmb0)
        >>> (1. * u.mK).to(u.MJy / u.sr, equivalencies=equiv)  # doctest: +FLOAT_CMP
        <Quantity 0.37993172 MJy / sr>

    """
    nu = frequency.to(si.GHz, spectral())

    if T_cmb is None:
        from astropy.cosmology import default_cosmology

        T_cmb = default_cosmology.get().Tcmb0

    def f(nu, T_cmb=T_cmb):
        x = _si.h * nu / _si.k_B / T_cmb
        return x**2 * np.exp(x) / np.expm1(x) ** 2

    def convert_Jy_to_K(x_jybm):
        factor = (f(nu) * 2 * _si.k_B * si.K * nu**2 / _si.c**2).to_value(astrophys.Jy)
        return x_jybm / factor

    def convert_K_to_Jy(x_K):
        factor = (astrophys.Jy / (f(nu) * 2 * _si.k_B * nu**2 / _si.c**2)).to_value(
            si.K
        )
        return x_K / factor

    return Equivalency(
        [(astrophys.Jy / si.sr, si.K, convert_Jy_to_K, convert_K_to_Jy)],
        "thermodynamic_temperature",
        {"frequency": frequency, "T_cmb": T_cmb},
    )


@functools.cache
def temperature():
    """Convert degrees Celsius and degrees Fahrenheit here because
    Unit and CompositeUnit cannot do addition or subtraction properly.
    """
    from .imperial import deg_F as F

    K = si.K
    C = si.deg_C

    return Equivalency(
        [
            (K, C, lambda x: x - 273.15, lambda x: x + 273.15),
            (C, F, lambda x: x * 1.8 + 32.0, lambda x: (x - 32.0) / 1.8),
            (K, F, lambda x: x * 1.8 - 459.67, lambda x: (x + 459.67) / 1.8),
        ],
        "temperature",
    )


@functools.cache
def temperature_energy():
    """Convert between Kelvin and keV(eV) to an equivalent amount."""
    e = _si.e.value
    k_B = _si.k_B.value
    return Equivalency(
        [(si.K, misc.eV, lambda x: x / (e / k_B), lambda x: x * (e / k_B))],
        "temperature_energy",
    )


def assert_is_spectral_unit(value):
    try:
        value.to(si.Hz, spectral())
    except (AttributeError, UnitsError) as ex:
        raise UnitsError(
            "The 'rest' value must be a spectral equivalent "
            "(frequency, wavelength, or energy)."
        )


def pixel_scale(pixscale):
    """
    Convert between pixel distances (in units of ``pix``) and other units,
    given a particular ``pixscale``.

    Parameters
    ----------
    pixscale : `~astropy.units.Quantity`
        The pixel scale either in units of <unit>/pixel or pixel/<unit>.
    """
    decomposed = pixscale.unit.decompose()
    dimensions = dict(zip(decomposed.bases, decomposed.powers))
    pix_power = dimensions.get(misc.pix, 0)

    if pix_power == -1:
        physical_unit = Unit(pixscale * misc.pix)
    elif pix_power == 1:
        physical_unit = Unit(misc.pix / pixscale)
    else:
        raise UnitsError(
            "The pixel scale unit must have pixel dimensionality of 1 or -1."
        )

    return Equivalency(
        [(misc.pix, physical_unit)], "pixel_scale", {"pixscale": pixscale}
    )


def plate_scale(platescale):
    """
    Convert between lengths (to be interpreted as lengths in the focal plane)
    and angular units with a specified ``platescale``.

    Parameters
    ----------
    platescale : `~astropy.units.Quantity`
        The pixel scale either in units of distance/pixel or distance/angle.
    """
    if platescale.unit.is_equivalent(si.arcsec / si.m):
        platescale_val = platescale.to_value(si.radian / si.m)
    elif platescale.unit.is_equivalent(si.m / si.arcsec):
        platescale_val = (1 / platescale).to_value(si.radian / si.m)
    else:
        raise UnitsError("The pixel scale must be in angle/distance or distance/angle")

    return Equivalency(
        [(si.m, si.radian, lambda d: d * platescale_val, lambda a: a / platescale_val)],
        "plate_scale",
        {"platescale": platescale},
    )


def magnetic_flux_field(mu_r=1):
    r"""
    Convert magnetic field between magnetic field strength :math:`(\mathbf{H})` and
    magnetic flux density :math:`(\mathbf{B})` using the relationship:

    .. math::

        \mathbf{B} = \mu_r \mu_0 \mathbf{H}

    where:
        - :math:`\mu_0` is the vacuum permeability, a physical constant.
        - :math:`\mu_r` is the relative permeability of the medium, a dimensionless
          quantity.

    The default setting (:math:`\mu_r=1`) represents conditions in a vacuum.

    Parameters
    ----------
    mu_r : float, optional
        The relative magnetic permeability of the medium. This is a dimensionless quantity
        and has a default value of :math:`\mu_r=1` which corresponds to free space (vacuum).

    Examples
    --------
    >>> import astropy.units as u
    >>> H = 1 * u.Oe
    >>> H.to(u.G, equivalencies=u.magnetic_flux_field())  # doctest: +FLOAT_CMP
    <Quantity 1. G>
    >>> H.to(u.G, equivalencies=u.magnetic_flux_field(mu_r=0.8))  # doctest: +FLOAT_CMP
    <Quantity 0.8 G>
    >>> B = 1 * u.T
    >>> B.to(u.A / u.m, equivalencies=u.magnetic_flux_field())  # doctest: +FLOAT_CMP
    <Quantity 795774.71502628 A / m>
    >>> B.to(u.A / u.m, equivalencies=u.magnetic_flux_field(mu_r=0.8))  # doctest: +FLOAT_CMP
    <Quantity 994718.39378285 A / m>

    """
    mu0 = _si.mu0.value
    return Equivalency(
        [(si.T, si.A / si.m, lambda x: x / (mu_r * mu0), lambda x: x * mu_r * mu0)],
        "magnetic_flux_field",
    )


def zero_point_flux(flux0):
    """
    An equivalency for converting linear flux units ("maggys") defined relative
    to a standard source into a standardized system.

    Parameters
    ----------
    flux0 : `~astropy.units.Quantity`
        The flux of a magnitude-0 object in the "maggy" system.
    """
    return Equivalency([(maggy, Unit(flux0))], "zero_point_flux")



# This stub file was automatically generated.

__all__ = [
    "Bol",
    "L_bol",
    "bol",
    "f_bol",
    "AB",
    "ABflux",
    "ST",
    "STflux",
    "mgy",
    "maggy",
    "nmgy",
    "nanomaggy",
]

from astropy.units.core import IrreducibleUnit
from astropy.units.core import PrefixUnit
from astropy.units.core import Unit

Bol: Unit
L_bol: Unit
bol: Unit
f_bol: Unit
AB: Unit
ABflux: Unit
ST: Unit
STflux: Unit
mgy: IrreducibleUnit
maggy: IrreducibleUnit
nmgy: PrefixUnit
nanomaggy: PrefixUnit



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import operator
from copy import deepcopy

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from astropy import units as u
from astropy.nddata import CCDData, NDDataRef
from astropy.nddata import _testing as nd_testing
from astropy.nddata.nduncertainty import (
    IncompatibleUncertaintiesException,
    InverseVariance,
    StdDevUncertainty,
    UnknownUncertainty,
    VarianceUncertainty,
)
from astropy.units import Quantity, UnitsError
from astropy.utils.compat import NUMPY_LT_2_0
from astropy.utils.exceptions import AstropyUserWarning
from astropy.wcs import WCS

# Alias NDDataAllMixins in case this will be renamed ... :-)
NDDataArithmetic = NDDataRef


class StdDevUncertaintyUncorrelated(StdDevUncertainty):
    @property
    def supports_correlated(self):
        return False


# Correspondence between NDArithmetic & Python method/function names:
STR_TO_OPERATOR = {
    "add": operator.add,
    "subtract": operator.sub,
    "multiply": operator.mul,
    "divide": operator.truediv,
}


# Test with Data covers:
# scalars, 1D, 2D and 3D
# broadcasting between them
@pytest.mark.filterwarnings("ignore:divide by zero encountered.*")
@pytest.mark.parametrize(
    ("data1", "data2"),
    [
        (np.array(5), np.array(10)),
        (np.array(5), np.arange(10)),
        (np.array(5), np.arange(10).reshape(2, 5)),
        (np.arange(10), np.ones(10) * 2),
        (np.arange(10), np.ones((10, 10)) * 2),
        (np.arange(10).reshape(2, 5), np.ones((2, 5)) * 3),
        (np.arange(1000).reshape(20, 5, 10), np.ones((20, 5, 10)) * 3),
    ],
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_data(data1, data2, meth):
    nd1 = NDDataArithmetic(data1)
    nd2 = NDDataArithmetic(data2)

    nd = getattr(nd1, meth)(nd2)
    assert_array_equal(STR_TO_OPERATOR[meth](data1, data2), nd.data)

    # Check that broadcasting worked as expected
    if data1.ndim > data2.ndim:
        assert data1.shape == nd.data.shape
    else:
        assert data2.shape == nd.data.shape
    # Check all other attributes are not set
    assert nd.unit is None
    assert nd.uncertainty is None
    assert nd.mask is None
    assert len(nd.meta) == 0
    assert nd.wcs is None


# Test numpy functions that use astropy functions first
def test_arithmetics_ccddata():
    ccd1 = CCDData([1, 2, 3], unit="adu")
    ccd2 = CCDData([1.1, 2.2, 3.3], unit="adu")
    ccd3 = CCDData([1.1, 2.2, 3.3, 4.4], unit="adu")
    nd1 = NDDataArithmetic(ccd1)
    nd3 = NDDataArithmetic(ccd3)

    assert np.min(ccd1).data == ccd1.min().data
    assert np.min(ccd2).data == ccd2.min().data

    assert np.max(ccd1).data == ccd1.max().data
    assert np.max(ccd2).data == ccd2.max().data

    assert np.sum(ccd1).data == ccd1.sum().data
    assert np.sum(ccd2).data == ccd2.sum().data

    assert np.mean(ccd1).data == ccd1.mean().data
    assert np.mean(ccd2).data == ccd2.mean().data


# Ensure exceptions are raised for numpy keys that are not None
def test_arithmetics_ccddata_errors():
    ccd1 = CCDData([1, 2, 3], unit="adu")
    ccd2 = CCDData([1.1, 2.2, 3.3, 4.4], unit="adu")
    nd1 = NDDataArithmetic(ccd1)
    nd2 = NDDataArithmetic(ccd2)

    with pytest.raises(ValueError):
        np.mean(ccd1, out=nd1)

    with pytest.raises(ValueError):
        np.mean(ccd2, out=nd2)

    with pytest.raises(ValueError):
        np.mean(ccd1, out=nd1, dtype=int)

    with pytest.raises(ValueError):
        np.mean(ccd2, dtype=int)


# Invalid arithmetic operations for data covering:
# not broadcastable data
def test_arithmetics_data_invalid():
    nd1 = NDDataArithmetic([1, 2, 3])
    nd2 = NDDataArithmetic([1, 2])
    with pytest.raises(ValueError):
        nd1.add(nd2)


# Test with Data and unit and covers:
# identical units (even dimensionless unscaled vs. no unit),
# equivalent units (such as meter and kilometer)
# equivalent composite units (such as m/s and km/h)
@pytest.mark.filterwarnings("ignore:divide by zero encountered.*")
@pytest.mark.parametrize(
    ("data1", "data2"),
    [
        (np.array(5) * u.s, np.array(10) * u.s),
        (np.array(5) * u.s, np.arange(10) * u.h),
        (np.array(5) * u.s, np.arange(10).reshape(2, 5) * u.min),
        (np.arange(10) * u.m / u.s, np.ones(10) * 2 * u.km / u.s),
        (np.arange(10) * u.m / u.s, np.ones((10, 10)) * 2 * u.m / u.h),
        (np.arange(10).reshape(2, 5) * u.m / u.s, np.ones((2, 5)) * 3 * u.km / u.h),
        (
            np.arange(1000).reshape(20, 5, 10),
            np.ones((20, 5, 10)) * 3 * u.dimensionless_unscaled,
        ),
        (np.array(5), np.array(10) * u.s / u.h),
    ],
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_data_unit_identical(data1, data2, meth):
    nd1 = NDDataArithmetic(data1)
    nd2 = NDDataArithmetic(data2)

    nd = getattr(nd1, meth)(nd2)
    ref = STR_TO_OPERATOR[meth](data1, data2)
    ref_unit, ref_data = ref.unit, ref.value
    assert_array_equal(ref_data, nd.data)
    assert nd.unit == ref_unit

    # Check that broadcasting worked as expected
    if data1.ndim > data2.ndim:
        assert data1.shape == nd.data.shape
    else:
        assert data2.shape == nd.data.shape
    # Check all other attributes are not set
    assert nd.uncertainty is None
    assert nd.mask is None
    assert len(nd.meta) == 0
    assert nd.wcs is None


# Test with Data and unit and covers:
# not identical not convertible units
# one with unit (which is not dimensionless) and one without
@pytest.mark.parametrize(
    ("data1", "data2"),
    [
        (np.array(5) * u.s, np.array(10) * u.m),
        (np.array(5) * u.Mpc, np.array(10) * u.km / u.s),
        (np.array(5) * u.Mpc, np.array(10)),
        (np.array(5), np.array(10) * u.s),
    ],
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_data_unit_not_identical(data1, data2, meth):
    nd1 = NDDataArithmetic(data1)
    nd2 = NDDataArithmetic(data2)

    if meth in ("add", "subtract"):
        # Addition/subtraction should not be possible
        with pytest.raises(UnitsError):
            getattr(nd1, meth)(nd2)
    else:
        # Multiplication/division is possible
        nd = getattr(nd1, meth)(nd2)
        ref = STR_TO_OPERATOR[meth](data1, data2)
        ref_unit, ref_data = ref.unit, ref.value
        assert_array_equal(ref_data, nd.data)
        assert nd.unit == ref_unit

        # Check all other attributes are not set
        assert nd.uncertainty is None
        assert nd.mask is None
        assert len(nd.meta) == 0
        assert nd.wcs is None


# Tests with wcs (not very sensible because there is no operation between them
# covering:
# both set and identical/not identical
# one set
# None set
@pytest.mark.parametrize(
    ("wcs1", "wcs2"),
    [
        (None, None),
        (None, WCS(naxis=2)),
        (WCS(naxis=2), None),
        nd_testing.create_two_equal_wcs(naxis=2),
        nd_testing.create_two_unequal_wcs(naxis=2),
    ],
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_data_wcs(wcs1, wcs2, meth):
    nd1 = NDDataArithmetic(1, wcs=wcs1)
    nd2 = NDDataArithmetic(1, wcs=wcs2)

    if wcs1 is None and wcs2 is None:
        ref_wcs = None
    elif wcs1 is None:
        ref_wcs = wcs2
    elif wcs2 is None:
        ref_wcs = wcs1
    else:
        ref_wcs = wcs1

    nd = getattr(nd1, meth)(nd2)
    nd_testing.assert_wcs_seem_equal(ref_wcs, nd.wcs)

    # Check all other attributes are not set
    assert nd.unit is None
    assert nd.uncertainty is None
    assert len(nd.meta) == 0
    assert nd.mask is None


# Masks are completely separated in the NDArithmetics from the data so we need
# no correlated tests but covering:
# masks 1D, 2D and mixed cases with broadcasting
@pytest.mark.parametrize(
    ("mask1", "mask2"),
    [
        (None, None),
        (None, False),
        (True, None),
        (False, False),
        (True, False),
        (False, True),
        (True, True),
        (np.array(False), np.array(True)),
        (np.array(False), np.array([0, 1, 0, 1, 1], dtype=np.bool_)),
        (np.array(True), np.array([[0, 1, 0, 1, 1], [1, 1, 0, 1, 1]], dtype=np.bool_)),
        (
            np.array([0, 1, 0, 1, 1], dtype=np.bool_),
            np.array([1, 1, 0, 0, 1], dtype=np.bool_),
        ),
        (
            np.array([0, 1, 0, 1, 1], dtype=np.bool_),
            np.array([[0, 1, 0, 1, 1], [1, 0, 0, 1, 1]], dtype=np.bool_),
        ),
        (
            np.array([[0, 1, 0, 1, 1], [1, 0, 0, 1, 1]], dtype=np.bool_),
            np.array([[0, 1, 0, 1, 1], [1, 1, 0, 1, 1]], dtype=np.bool_),
        ),
    ],
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_data_masks(mask1, mask2, meth):
    nd1 = NDDataArithmetic(1, mask=mask1)
    nd2 = NDDataArithmetic(1, mask=mask2)

    if mask1 is None and mask2 is None:
        ref_mask = None
    elif mask1 is None:
        ref_mask = mask2
    elif mask2 is None:
        ref_mask = mask1
    else:
        ref_mask = mask1 | mask2

    nd = getattr(nd1, meth)(nd2)
    assert_array_equal(ref_mask, nd.mask)

    # Check all other attributes are not set
    assert nd.unit is None
    assert nd.uncertainty is None
    assert len(nd.meta) == 0
    assert nd.wcs is None


# Check that masks are preserved+propagated in NDData collapse operations
@pytest.mark.parametrize(
    ("collapse_axis", "mask_sum", "unit"),
    [(0, [3, 0, 3, 0], "Jy"), (1, [2, 0, 2, 0], None), (2, [2, 2, 2], "Jy")],
)
def test_collapse_masks(collapse_axis, mask_sum, unit):
    shape = (2, 3, 4)
    data = np.arange(np.prod(shape)).reshape(shape)
    mask = data % 2 == 0
    nd_masked = NDDataArithmetic(data=data, mask=mask, unit=unit)
    nd_nomask = NDDataArithmetic(data=data, unit=unit)

    assert_array_equal(nd_masked.sum(axis=collapse_axis).mask.sum(axis=0), mask_sum)

    # if no mask is given, the collapse result should have no mask:
    assert nd_nomask.sum(axis=collapse_axis).mask is None


# One additional case which can not be easily incorporated in the test above
# what happens if the masks are numpy ndarrays are not broadcastable
def test_arithmetics_data_masks_invalid():
    nd1 = NDDataArithmetic(1, mask=np.array([1, 0], dtype=np.bool_))
    nd2 = NDDataArithmetic(1, mask=np.array([1, 0, 1], dtype=np.bool_))

    with pytest.raises(ValueError):
        nd1.add(nd2)
    with pytest.raises(ValueError):
        nd1.multiply(nd2)
    with pytest.raises(ValueError):
        nd1.subtract(nd2)
    with pytest.raises(ValueError):
        nd1.divide(nd2)


# Covering:
# both have uncertainties (data and uncertainty without unit)
# tested against manually determined resulting uncertainties to verify the
# implemented formulas
# this test only works as long as data1 and data2 do not contain any 0
def test_arithmetics_stddevuncertainty_basic():
    nd1 = NDDataArithmetic([1, 2, 3], uncertainty=StdDevUncertainty([1, 1, 3]))
    nd2 = NDDataArithmetic([2, 2, 2], uncertainty=StdDevUncertainty([2, 2, 2]))
    nd3 = nd1.add(nd2)
    nd4 = nd2.add(nd1)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = np.sqrt(np.array([1, 1, 3]) ** 2 + np.array([2, 2, 2]) ** 2)
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.subtract(nd2)
    nd4 = nd2.subtract(nd1)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty (same as for add)
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    # Multiplication and Division only work with almost equal array comparisons
    # since the formula implemented and the formula used as reference are
    # slightly different.
    nd3 = nd1.multiply(nd2)
    nd4 = nd2.multiply(nd1)
    # Inverse operation should result in the same uncertainty
    assert_array_almost_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = np.abs(np.array([2, 4, 6])) * np.sqrt(
        (np.array([1, 1, 3]) / np.array([1, 2, 3])) ** 2
        + (np.array([2, 2, 2]) / np.array([2, 2, 2])) ** 2
    )
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.divide(nd2)
    nd4 = nd2.divide(nd1)
    # Inverse operation gives a different uncertainty!
    # Compare it to the theoretical uncertainty
    ref_uncertainty_1 = np.abs(np.array([1 / 2, 2 / 2, 3 / 2])) * np.sqrt(
        (np.array([1, 1, 3]) / np.array([1, 2, 3])) ** 2
        + (np.array([2, 2, 2]) / np.array([2, 2, 2])) ** 2
    )
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty_1)
    ref_uncertainty_2 = np.abs(np.array([2, 1, 2 / 3])) * np.sqrt(
        (np.array([1, 1, 3]) / np.array([1, 2, 3])) ** 2
        + (np.array([2, 2, 2]) / np.array([2, 2, 2])) ** 2
    )
    assert_array_almost_equal(nd4.uncertainty.array, ref_uncertainty_2)


# Tests for correlation, covering
# correlation between -1 and 1 with correlation term being positive / negative
# also with one data being once positive and once completely negative
# The point of this test is to compare the used formula to the theoretical one.
# TODO: Maybe covering units too but I think that should work because of
# the next tests. Also this may be reduced somehow.
@pytest.mark.parametrize(
    ("cor", "uncert1", "data2"),
    [
        (-1, [1, 1, 3], [2, 2, 7]),
        (-0.5, [1, 1, 3], [2, 2, 7]),
        (-0.25, [1, 1, 3], [2, 2, 7]),
        (0, [1, 1, 3], [2, 2, 7]),
        (0.25, [1, 1, 3], [2, 2, 7]),
        (0.5, [1, 1, 3], [2, 2, 7]),
        (1, [1, 1, 3], [2, 2, 7]),
        (-1, [-1, -1, -3], [2, 2, 7]),
        (-0.5, [-1, -1, -3], [2, 2, 7]),
        (-0.25, [-1, -1, -3], [2, 2, 7]),
        (0, [-1, -1, -3], [2, 2, 7]),
        (0.25, [-1, -1, -3], [2, 2, 7]),
        (0.5, [-1, -1, -3], [2, 2, 7]),
        (1, [-1, -1, -3], [2, 2, 7]),
        (-1, [1, 1, 3], [-2, -3, -2]),
        (-0.5, [1, 1, 3], [-2, -3, -2]),
        (-0.25, [1, 1, 3], [-2, -3, -2]),
        (0, [1, 1, 3], [-2, -3, -2]),
        (0.25, [1, 1, 3], [-2, -3, -2]),
        (0.5, [1, 1, 3], [-2, -3, -2]),
        (1, [1, 1, 3], [-2, -3, -2]),
        (-1, [-1, -1, -3], [-2, -3, -2]),
        (-0.5, [-1, -1, -3], [-2, -3, -2]),
        (-0.25, [-1, -1, -3], [-2, -3, -2]),
        (0, [-1, -1, -3], [-2, -3, -2]),
        (0.25, [-1, -1, -3], [-2, -3, -2]),
        (0.5, [-1, -1, -3], [-2, -3, -2]),
        (1, [-1, -1, -3], [-2, -3, -2]),
    ],
)
def test_arithmetics_stddevuncertainty_basic_with_correlation(cor, uncert1, data2):
    data1 = np.array([1, 2, 3])
    data2 = np.array(data2)
    uncert1 = np.array(uncert1)
    uncert2 = np.array([2, 2, 2])
    nd1 = NDDataArithmetic(data1, uncertainty=StdDevUncertainty(uncert1))
    nd2 = NDDataArithmetic(data2, uncertainty=StdDevUncertainty(uncert2))
    nd3 = nd1.add(nd2, uncertainty_correlation=cor)
    nd4 = nd2.add(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = np.sqrt(
        uncert1**2 + uncert2**2 + 2 * cor * np.abs(uncert1 * uncert2)
    )
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.subtract(nd2, uncertainty_correlation=cor)
    nd4 = nd2.subtract(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = np.sqrt(
        uncert1**2 + uncert2**2 - 2 * cor * np.abs(uncert1 * uncert2)
    )
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    # Multiplication and Division only work with almost equal array comparisons
    # since the formula implemented and the formula used as reference are
    # slightly different.
    nd3 = nd1.multiply(nd2, uncertainty_correlation=cor)
    nd4 = nd2.multiply(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_almost_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = (np.abs(data1 * data2)) * np.sqrt(
        (uncert1 / data1) ** 2
        + (uncert2 / data2) ** 2
        + (2 * cor * np.abs(uncert1 * uncert2) / (data1 * data2))
    )
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.divide(nd2, uncertainty_correlation=cor)
    nd4 = nd2.divide(nd1, uncertainty_correlation=cor)
    # Inverse operation gives a different uncertainty!
    # Compare it to the theoretical uncertainty
    ref_uncertainty_1 = (np.abs(data1 / data2)) * np.sqrt(
        (uncert1 / data1) ** 2
        + (uncert2 / data2) ** 2
        - (2 * cor * np.abs(uncert1 * uncert2) / (data1 * data2))
    )
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty_1)
    ref_uncertainty_2 = (np.abs(data2 / data1)) * np.sqrt(
        (uncert1 / data1) ** 2
        + (uncert2 / data2) ** 2
        - (2 * cor * np.abs(uncert1 * uncert2) / (data1 * data2))
    )
    assert_array_almost_equal(nd4.uncertainty.array, ref_uncertainty_2)


# Tests for correlation, covering
# correlation between -1 and 1 with correlation term being positive / negative
# also with one data being once positive and once completely negative
# The point of this test is to compare the used formula to the theoretical one.
# TODO: Maybe covering units too but I think that should work because of
# the next tests. Also this may be reduced somehow.
@pytest.mark.parametrize(
    ("cor", "uncert1", "data2"),
    [
        (-1, [1, 1, 3], [2, 2, 7]),
        (-0.5, [1, 1, 3], [2, 2, 7]),
        (-0.25, [1, 1, 3], [2, 2, 7]),
        (0, [1, 1, 3], [2, 2, 7]),
        (0.25, [1, 1, 3], [2, 2, 7]),
        (0.5, [1, 1, 3], [2, 2, 7]),
        (1, [1, 1, 3], [2, 2, 7]),
        (-1, [-1, -1, -3], [2, 2, 7]),
        (-0.5, [-1, -1, -3], [2, 2, 7]),
        (-0.25, [-1, -1, -3], [2, 2, 7]),
        (0, [-1, -1, -3], [2, 2, 7]),
        (0.25, [-1, -1, -3], [2, 2, 7]),
        (0.5, [-1, -1, -3], [2, 2, 7]),
        (1, [-1, -1, -3], [2, 2, 7]),
        (-1, [1, 1, 3], [-2, -3, -2]),
        (-0.5, [1, 1, 3], [-2, -3, -2]),
        (-0.25, [1, 1, 3], [-2, -3, -2]),
        (0, [1, 1, 3], [-2, -3, -2]),
        (0.25, [1, 1, 3], [-2, -3, -2]),
        (0.5, [1, 1, 3], [-2, -3, -2]),
        (1, [1, 1, 3], [-2, -3, -2]),
        (-1, [-1, -1, -3], [-2, -3, -2]),
        (-0.5, [-1, -1, -3], [-2, -3, -2]),
        (-0.25, [-1, -1, -3], [-2, -3, -2]),
        (0, [-1, -1, -3], [-2, -3, -2]),
        (0.25, [-1, -1, -3], [-2, -3, -2]),
        (0.5, [-1, -1, -3], [-2, -3, -2]),
        (1, [-1, -1, -3], [-2, -3, -2]),
    ],
)
def test_arithmetics_varianceuncertainty_basic_with_correlation(cor, uncert1, data2):
    data1 = np.array([1, 2, 3])
    data2 = np.array(data2)
    uncert1 = np.array(uncert1) ** 2
    uncert2 = np.array([2, 2, 2]) ** 2
    nd1 = NDDataArithmetic(data1, uncertainty=VarianceUncertainty(uncert1))
    nd2 = NDDataArithmetic(data2, uncertainty=VarianceUncertainty(uncert2))
    nd3 = nd1.add(nd2, uncertainty_correlation=cor)
    nd4 = nd2.add(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = uncert1 + uncert2 + 2 * cor * np.sqrt(uncert1 * uncert2)
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.subtract(nd2, uncertainty_correlation=cor)
    nd4 = nd2.subtract(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = uncert1 + uncert2 - 2 * cor * np.sqrt(uncert1 * uncert2)
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    # Multiplication and Division only work with almost equal array comparisons
    # since the formula implemented and the formula used as reference are
    # slightly different.
    nd3 = nd1.multiply(nd2, uncertainty_correlation=cor)
    nd4 = nd2.multiply(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_almost_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = (data1 * data2) ** 2 * (
        uncert1 / data1**2
        + uncert2 / data2**2
        + (2 * cor * np.sqrt(uncert1 * uncert2) / (data1 * data2))
    )
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.divide(nd2, uncertainty_correlation=cor)
    nd4 = nd2.divide(nd1, uncertainty_correlation=cor)
    # Inverse operation gives a different uncertainty because of the
    # prefactor nd1/nd2 vs nd2/nd1. Howeveare, a large chunk is the same.
    ref_common = (
        uncert1 / data1**2
        + uncert2 / data2**2
        - (2 * cor * np.sqrt(uncert1 * uncert2) / (data1 * data2))
    )
    # Compare it to the theoretical uncertainty
    ref_uncertainty_1 = (data1 / data2) ** 2 * ref_common
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty_1)
    ref_uncertainty_2 = (data2 / data1) ** 2 * ref_common
    assert_array_almost_equal(nd4.uncertainty.array, ref_uncertainty_2)


# Tests for correlation, covering
# correlation between -1 and 1 with correlation term being positive / negative
# also with one data being once positive and once completely negative
# The point of this test is to compare the used formula to the theoretical one.
# TODO: Maybe covering units too but I think that should work because of
# the next tests. Also this may be reduced somehow.
@pytest.mark.filterwarnings("ignore:divide by zero encountered.*")
@pytest.mark.parametrize(
    ("cor", "uncert1", "data2"),
    [
        (-1, [1, 1, 3], [2, 2, 7]),
        (-0.5, [1, 1, 3], [2, 2, 7]),
        (-0.25, [1, 1, 3], [2, 2, 7]),
        (0, [1, 1, 3], [2, 2, 7]),
        (0.25, [1, 1, 3], [2, 2, 7]),
        (0.5, [1, 1, 3], [2, 2, 7]),
        (1, [1, 1, 3], [2, 2, 7]),
        (-1, [-1, -1, -3], [2, 2, 7]),
        (-0.5, [-1, -1, -3], [2, 2, 7]),
        (-0.25, [-1, -1, -3], [2, 2, 7]),
        (0, [-1, -1, -3], [2, 2, 7]),
        (0.25, [-1, -1, -3], [2, 2, 7]),
        (0.5, [-1, -1, -3], [2, 2, 7]),
        (1, [-1, -1, -3], [2, 2, 7]),
        (-1, [1, 1, 3], [-2, -3, -2]),
        (-0.5, [1, 1, 3], [-2, -3, -2]),
        (-0.25, [1, 1, 3], [-2, -3, -2]),
        (0, [1, 1, 3], [-2, -3, -2]),
        (0.25, [1, 1, 3], [-2, -3, -2]),
        (0.5, [1, 1, 3], [-2, -3, -2]),
        (1, [1, 1, 3], [-2, -3, -2]),
        (-1, [-1, -1, -3], [-2, -3, -2]),
        (-0.5, [-1, -1, -3], [-2, -3, -2]),
        (-0.25, [-1, -1, -3], [-2, -3, -2]),
        (0, [-1, -1, -3], [-2, -3, -2]),
        (0.25, [-1, -1, -3], [-2, -3, -2]),
        (0.5, [-1, -1, -3], [-2, -3, -2]),
        (1, [-1, -1, -3], [-2, -3, -2]),
    ],
)
def test_arithmetics_inversevarianceuncertainty_basic_with_correlation(
    cor, uncert1, data2
):
    data1 = np.array([1, 2, 3])
    data2 = np.array(data2)
    uncert1 = 1 / np.array(uncert1) ** 2
    uncert2 = 1 / np.array([2, 2, 2]) ** 2
    nd1 = NDDataArithmetic(data1, uncertainty=InverseVariance(uncert1))
    nd2 = NDDataArithmetic(data2, uncertainty=InverseVariance(uncert2))
    nd3 = nd1.add(nd2, uncertainty_correlation=cor)
    nd4 = nd2.add(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = 1 / (
        1 / uncert1 + 1 / uncert2 + 2 * cor / np.sqrt(uncert1 * uncert2)
    )
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.subtract(nd2, uncertainty_correlation=cor)
    nd4 = nd2.subtract(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = 1 / (
        1 / uncert1 + 1 / uncert2 - 2 * cor / np.sqrt(uncert1 * uncert2)
    )
    assert_array_equal(nd3.uncertainty.array, ref_uncertainty)

    # Multiplication and Division only work with almost equal array comparisons
    # since the formula implemented and the formula used as reference are
    # slightly different.
    nd3 = nd1.multiply(nd2, uncertainty_correlation=cor)
    nd4 = nd2.multiply(nd1, uncertainty_correlation=cor)
    # Inverse operation should result in the same uncertainty
    assert_array_almost_equal(nd3.uncertainty.array, nd4.uncertainty.array)
    # Compare it to the theoretical uncertainty
    ref_uncertainty = 1 / (
        (data1 * data2) ** 2
        * (
            1 / uncert1 / data1**2
            + 1 / uncert2 / data2**2
            + (2 * cor / np.sqrt(uncert1 * uncert2) / (data1 * data2))
        )
    )
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty)

    nd3 = nd1.divide(nd2, uncertainty_correlation=cor)
    nd4 = nd2.divide(nd1, uncertainty_correlation=cor)
    # Inverse operation gives a different uncertainty because of the
    # prefactor nd1/nd2 vs nd2/nd1. Howeveare, a large chunk is the same.
    ref_common = (
        1 / uncert1 / data1**2
        + 1 / uncert2 / data2**2
        - (2 * cor / np.sqrt(uncert1 * uncert2) / (data1 * data2))
    )
    # Compare it to the theoretical uncertainty
    ref_uncertainty_1 = 1 / ((data1 / data2) ** 2 * ref_common)
    assert_array_almost_equal(nd3.uncertainty.array, ref_uncertainty_1)
    ref_uncertainty_2 = 1 / ((data2 / data1) ** 2 * ref_common)
    assert_array_almost_equal(nd4.uncertainty.array, ref_uncertainty_2)


# Covering:
# just an example that a np.ndarray works as correlation, no checks for
# the right result since these were basically done in the function above.
def test_arithmetics_stddevuncertainty_basic_with_correlation_array():
    data1 = np.array([1, 2, 3])
    data2 = np.array([1, 1, 1])
    uncert1 = np.array([1, 1, 1])
    uncert2 = np.array([2, 2, 2])
    cor = np.array([0, 0.25, 0])
    nd1 = NDDataArithmetic(data1, uncertainty=StdDevUncertainty(uncert1))
    nd2 = NDDataArithmetic(data2, uncertainty=StdDevUncertainty(uncert2))
    nd1.add(nd2, uncertainty_correlation=cor)


# Covering:
# That propagate throws an exception when correlation is given but the
# uncertainty does not support correlation.
def test_arithmetics_with_correlation_unsupported():
    data1 = np.array([1, 2, 3])
    data2 = np.array([1, 1, 1])
    uncert1 = np.array([1, 1, 1])
    uncert2 = np.array([2, 2, 2])
    cor = 3
    nd1 = NDDataArithmetic(data1, uncertainty=StdDevUncertaintyUncorrelated(uncert1))
    nd2 = NDDataArithmetic(data2, uncertainty=StdDevUncertaintyUncorrelated(uncert2))

    with pytest.raises(ValueError):
        nd1.add(nd2, uncertainty_correlation=cor)


# Covering:
# only one has an uncertainty (data and uncertainty without unit)
# tested against the case where the other one has zero uncertainty. (this case
# must be correct because we tested it in the last case)
# Also verify that if the result of the data has negative values the resulting
# uncertainty has no negative values.
def test_arithmetics_stddevuncertainty_one_missing():
    nd1 = NDDataArithmetic([1, -2, 3])
    nd1_ref = NDDataArithmetic([1, -2, 3], uncertainty=StdDevUncertainty([0, 0, 0]))
    nd2 = NDDataArithmetic([2, 2, -2], uncertainty=StdDevUncertainty([2, 2, 2]))

    # Addition
    nd3 = nd1.add(nd2)
    nd3_ref = nd1_ref.add(nd2)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    nd3 = nd2.add(nd1)
    nd3_ref = nd2.add(nd1_ref)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    # Subtraction
    nd3 = nd1.subtract(nd2)
    nd3_ref = nd1_ref.subtract(nd2)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    nd3 = nd2.subtract(nd1)
    nd3_ref = nd2.subtract(nd1_ref)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    # Multiplication
    nd3 = nd1.multiply(nd2)
    nd3_ref = nd1_ref.multiply(nd2)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    nd3 = nd2.multiply(nd1)
    nd3_ref = nd2.multiply(nd1_ref)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    # Division
    nd3 = nd1.divide(nd2)
    nd3_ref = nd1_ref.divide(nd2)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)

    nd3 = nd2.divide(nd1)
    nd3_ref = nd2.divide(nd1_ref)
    assert_array_equal(nd3.uncertainty.array, nd3_ref.uncertainty.array)
    assert_array_equal(np.abs(nd3.uncertainty.array), nd3.uncertainty.array)


# Covering:
# data with unit and uncertainty with unit (but equivalent units)
# compared against correctly scaled NDDatas
@pytest.mark.filterwarnings("ignore:.*encountered in.*divide.*")
@pytest.mark.parametrize(
    ("uncert1", "uncert2"),
    [
        (np.array([1, 2, 3]) * u.m, None),
        (np.array([1, 2, 3]) * u.cm, None),
        (None, np.array([1, 2, 3]) * u.m),
        (None, np.array([1, 2, 3]) * u.cm),
        (np.array([1, 2, 3]), np.array([2, 3, 4])),
        (np.array([1, 2, 3]) * u.m, np.array([2, 3, 4])),
        (np.array([1, 2, 3]), np.array([2, 3, 4])) * u.m,
        (np.array([1, 2, 3]) * u.m, np.array([2, 3, 4])) * u.m,
        (np.array([1, 2, 3]) * u.cm, np.array([2, 3, 4])),
        (np.array([1, 2, 3]), np.array([2, 3, 4])) * u.cm,
        (np.array([1, 2, 3]) * u.cm, np.array([2, 3, 4])) * u.cm,
        (np.array([1, 2, 3]) * u.km, np.array([2, 3, 4])) * u.cm,
    ],
)
def test_arithmetics_stddevuncertainty_with_units(uncert1, uncert2):
    # Data has same units
    data1 = np.array([1, 2, 3]) * u.m
    data2 = np.array([-4, 7, 0]) * u.m
    if uncert1 is not None:
        uncert1 = StdDevUncertainty(uncert1)
        if isinstance(uncert1, Quantity):
            uncert1_ref = uncert1.to_value(data1.unit)
        else:
            uncert1_ref = uncert1
        uncert_ref1 = StdDevUncertainty(uncert1_ref, copy=True)
    else:
        uncert1 = None
        uncert_ref1 = None

    if uncert2 is not None:
        uncert2 = StdDevUncertainty(uncert2)
        if isinstance(uncert2, Quantity):
            uncert2_ref = uncert2.to_value(data2.unit)
        else:
            uncert2_ref = uncert2
        uncert_ref2 = StdDevUncertainty(uncert2_ref, copy=True)
    else:
        uncert2 = None
        uncert_ref2 = None

    nd1 = NDDataArithmetic(data1, uncertainty=uncert1)
    nd2 = NDDataArithmetic(data2, uncertainty=uncert2)

    nd1_ref = NDDataArithmetic(data1, uncertainty=uncert_ref1)
    nd2_ref = NDDataArithmetic(data2, uncertainty=uncert_ref2)

    # Let's start the tests
    # Addition
    nd3 = nd1.add(nd2)
    nd3_ref = nd1_ref.add(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.add(nd1)
    nd3_ref = nd2_ref.add(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Subtraction
    nd3 = nd1.subtract(nd2)
    nd3_ref = nd1_ref.subtract(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.subtract(nd1)
    nd3_ref = nd2_ref.subtract(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Multiplication
    nd3 = nd1.multiply(nd2)
    nd3_ref = nd1_ref.multiply(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.multiply(nd1)
    nd3_ref = nd2_ref.multiply(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Division
    nd3 = nd1.divide(nd2)
    nd3_ref = nd1_ref.divide(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.divide(nd1)
    nd3_ref = nd2_ref.divide(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)


# Covering:
# data with unit and uncertainty with unit (but equivalent units)
# compared against correctly scaled NDDatas
@pytest.mark.filterwarnings("ignore:.*encountered in.*divide.*")
@pytest.mark.parametrize(
    ("uncert1", "uncert2"),
    [
        (np.array([1, 2, 3]) * u.m, None),
        (np.array([1, 2, 3]) * u.cm, None),
        (None, np.array([1, 2, 3]) * u.m),
        (None, np.array([1, 2, 3]) * u.cm),
        (np.array([1, 2, 3]), np.array([2, 3, 4])),
        (np.array([1, 2, 3]) * u.m, np.array([2, 3, 4])),
        (np.array([1, 2, 3]), np.array([2, 3, 4])) * u.m,
        (np.array([1, 2, 3]) * u.m, np.array([2, 3, 4])) * u.m,
        (np.array([1, 2, 3]) * u.cm, np.array([2, 3, 4])),
        (np.array([1, 2, 3]), np.array([2, 3, 4])) * u.cm,
        (np.array([1, 2, 3]) * u.cm, np.array([2, 3, 4])) * u.cm,
        (np.array([1, 2, 3]) * u.km, np.array([2, 3, 4])) * u.cm,
    ],
)
def test_arithmetics_varianceuncertainty_with_units(uncert1, uncert2):
    # Data has same units
    data1 = np.array([1, 2, 3]) * u.m
    data2 = np.array([-4, 7, 0]) * u.m
    if uncert1 is not None:
        uncert1 = VarianceUncertainty(uncert1**2)
        if isinstance(uncert1, Quantity):
            uncert1_ref = uncert1.to_value(data1.unit**2)
        else:
            uncert1_ref = uncert1
        uncert_ref1 = VarianceUncertainty(uncert1_ref, copy=True)
    else:
        uncert1 = None
        uncert_ref1 = None

    if uncert2 is not None:
        uncert2 = VarianceUncertainty(uncert2**2)
        if isinstance(uncert2, Quantity):
            uncert2_ref = uncert2.to_value(data2.unit**2)
        else:
            uncert2_ref = uncert2
        uncert_ref2 = VarianceUncertainty(uncert2_ref, copy=True)
    else:
        uncert2 = None
        uncert_ref2 = None

    nd1 = NDDataArithmetic(data1, uncertainty=uncert1)
    nd2 = NDDataArithmetic(data2, uncertainty=uncert2)

    nd1_ref = NDDataArithmetic(data1, uncertainty=uncert_ref1)
    nd2_ref = NDDataArithmetic(data2, uncertainty=uncert_ref2)

    # Let's start the tests
    # Addition
    nd3 = nd1.add(nd2)
    nd3_ref = nd1_ref.add(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.add(nd1)
    nd3_ref = nd2_ref.add(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Subtraction
    nd3 = nd1.subtract(nd2)
    nd3_ref = nd1_ref.subtract(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.subtract(nd1)
    nd3_ref = nd2_ref.subtract(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Multiplication
    nd3 = nd1.multiply(nd2)
    nd3_ref = nd1_ref.multiply(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.multiply(nd1)
    nd3_ref = nd2_ref.multiply(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Division
    nd3 = nd1.divide(nd2)
    nd3_ref = nd1_ref.divide(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.divide(nd1)
    nd3_ref = nd2_ref.divide(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)


# Covering:
# data with unit and uncertainty with unit (but equivalent units)
# compared against correctly scaled NDDatas
@pytest.mark.filterwarnings("ignore:.*encountered in.*divide.*")
@pytest.mark.parametrize(
    ("uncert1", "uncert2"),
    [
        (np.array([1, 2, 3]) * u.m, None),
        (np.array([1, 2, 3]) * u.cm, None),
        (None, np.array([1, 2, 3]) * u.m),
        (None, np.array([1, 2, 3]) * u.cm),
        (np.array([1, 2, 3]), np.array([2, 3, 4])),
        (np.array([1, 2, 3]) * u.m, np.array([2, 3, 4])),
        (np.array([1, 2, 3]), np.array([2, 3, 4])) * u.m,
        (np.array([1, 2, 3]) * u.m, np.array([2, 3, 4])) * u.m,
        (np.array([1, 2, 3]) * u.cm, np.array([2, 3, 4])),
        (np.array([1, 2, 3]), np.array([2, 3, 4])) * u.cm,
        (np.array([1, 2, 3]) * u.cm, np.array([2, 3, 4])) * u.cm,
        (np.array([1, 2, 3]) * u.km, np.array([2, 3, 4])) * u.cm,
    ],
)
def test_arithmetics_inversevarianceuncertainty_with_units(uncert1, uncert2):
    # Data has same units
    data1 = np.array([1, 2, 3]) * u.m
    data2 = np.array([-4, 7, 0]) * u.m
    if uncert1 is not None:
        uncert1 = InverseVariance(1 / uncert1**2)
        if isinstance(uncert1, Quantity):
            uncert1_ref = uncert1.to_value(1 / data1.unit**2)
        else:
            uncert1_ref = uncert1
        uncert_ref1 = InverseVariance(uncert1_ref, copy=True)
    else:
        uncert1 = None
        uncert_ref1 = None

    if uncert2 is not None:
        uncert2 = InverseVariance(1 / uncert2**2)
        if isinstance(uncert2, Quantity):
            uncert2_ref = uncert2.to_value(1 / data2.unit**2)
        else:
            uncert2_ref = uncert2
        uncert_ref2 = InverseVariance(uncert2_ref, copy=True)
    else:
        uncert2 = None
        uncert_ref2 = None

    nd1 = NDDataArithmetic(data1, uncertainty=uncert1)
    nd2 = NDDataArithmetic(data2, uncertainty=uncert2)

    nd1_ref = NDDataArithmetic(data1, uncertainty=uncert_ref1)
    nd2_ref = NDDataArithmetic(data2, uncertainty=uncert_ref2)

    # Let's start the tests
    # Addition
    nd3 = nd1.add(nd2)
    nd3_ref = nd1_ref.add(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.add(nd1)
    nd3_ref = nd2_ref.add(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Subtraction
    nd3 = nd1.subtract(nd2)
    nd3_ref = nd1_ref.subtract(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.subtract(nd1)
    nd3_ref = nd2_ref.subtract(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Multiplication
    nd3 = nd1.multiply(nd2)
    nd3_ref = nd1_ref.multiply(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.multiply(nd1)
    nd3_ref = nd2_ref.multiply(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    # Division
    nd3 = nd1.divide(nd2)
    nd3_ref = nd1_ref.divide(nd2_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)

    nd3 = nd2.divide(nd1)
    nd3_ref = nd2_ref.divide(nd1_ref)
    assert nd3.unit == nd3_ref.unit
    assert nd3.uncertainty.unit == nd3_ref.uncertainty.unit
    assert_array_equal(nd3.uncertainty.array, nd3.uncertainty.array)


# Test abbreviation and long name for taking the first found meta, mask, wcs
@pytest.mark.parametrize("use_abbreviation", ["ff", "first_found"])
def test_arithmetics_handle_switches(use_abbreviation):
    meta1 = {"a": 1}
    meta2 = {"b": 2}
    mask1 = True
    mask2 = False
    uncertainty1 = StdDevUncertainty([1, 2, 3])
    uncertainty2 = StdDevUncertainty([1, 2, 3])
    wcs1, wcs2 = nd_testing.create_two_unequal_wcs(naxis=1)
    data1 = [1, 1, 1]
    data2 = [1, 1, 1]

    nd1 = NDDataArithmetic(
        data1, meta=meta1, mask=mask1, wcs=wcs1, uncertainty=uncertainty1
    )
    nd2 = NDDataArithmetic(
        data2, meta=meta2, mask=mask2, wcs=wcs2, uncertainty=uncertainty2
    )
    nd3 = NDDataArithmetic(data1)

    # Both have the attributes but option None is chosen
    nd_ = nd1.add(
        nd2,
        propagate_uncertainties=None,
        handle_meta=None,
        handle_mask=None,
        compare_wcs=None,
    )
    assert nd_.wcs is None
    assert len(nd_.meta) == 0
    assert nd_.mask is None
    assert nd_.uncertainty is None

    # Only second has attributes and False is chosen
    nd_ = nd3.add(
        nd2,
        propagate_uncertainties=False,
        handle_meta=use_abbreviation,
        handle_mask=use_abbreviation,
        compare_wcs=use_abbreviation,
    )
    nd_testing.assert_wcs_seem_equal(nd_.wcs, wcs2)
    assert nd_.meta == meta2
    assert nd_.mask == mask2
    assert_array_equal(nd_.uncertainty.array, uncertainty2.array)

    # Only first has attributes and False is chosen
    nd_ = nd1.add(
        nd3,
        propagate_uncertainties=False,
        handle_meta=use_abbreviation,
        handle_mask=use_abbreviation,
        compare_wcs=use_abbreviation,
    )
    nd_testing.assert_wcs_seem_equal(nd_.wcs, wcs1)
    assert nd_.meta == meta1
    assert nd_.mask == mask1
    assert_array_equal(nd_.uncertainty.array, uncertainty1.array)


def test_arithmetics_meta_func():
    def meta_fun_func(meta1, meta2, take="first"):
        if take == "first":
            return meta1
        else:
            return meta2

    meta1 = {"a": 1}
    meta2 = {"a": 3, "b": 2}
    mask1 = True
    mask2 = False
    uncertainty1 = StdDevUncertainty([1, 2, 3])
    uncertainty2 = StdDevUncertainty([1, 2, 3])
    data1 = [1, 1, 1]
    data2 = [1, 1, 1]

    nd1 = NDDataArithmetic(data1, meta=meta1, mask=mask1, uncertainty=uncertainty1)
    nd2 = NDDataArithmetic(data2, meta=meta2, mask=mask2, uncertainty=uncertainty2)

    nd3 = nd1.add(nd2, handle_meta=meta_fun_func)
    assert nd3.meta["a"] == 1
    assert "b" not in nd3.meta

    nd4 = nd1.add(nd2, handle_meta=meta_fun_func, meta_take="second")
    assert nd4.meta["a"] == 3
    assert nd4.meta["b"] == 2

    with pytest.raises(KeyError):
        nd1.add(nd2, handle_meta=meta_fun_func, take="second")


def test_arithmetics_wcs_func():
    def wcs_comp_func(wcs1, wcs2, tolerance=0.1):
        if tolerance < 0.01:
            return False
        return True

    meta1 = {"a": 1}
    meta2 = {"a": 3, "b": 2}
    mask1 = True
    mask2 = False
    uncertainty1 = StdDevUncertainty([1, 2, 3])
    uncertainty2 = StdDevUncertainty([1, 2, 3])
    wcs1, wcs2 = nd_testing.create_two_equal_wcs(naxis=1)
    data1 = [1, 1, 1]
    data2 = [1, 1, 1]

    nd1 = NDDataArithmetic(
        data1, meta=meta1, mask=mask1, wcs=wcs1, uncertainty=uncertainty1
    )
    nd2 = NDDataArithmetic(
        data2, meta=meta2, mask=mask2, wcs=wcs2, uncertainty=uncertainty2
    )

    nd3 = nd1.add(nd2, compare_wcs=wcs_comp_func)
    nd_testing.assert_wcs_seem_equal(nd3.wcs, wcs1)

    # Fails because the function fails
    with pytest.raises(ValueError):
        nd1.add(nd2, compare_wcs=wcs_comp_func, wcs_tolerance=0.00001)

    # Fails because for a parameter to be passed correctly to the function it
    # needs the wcs_ prefix
    with pytest.raises(KeyError):
        nd1.add(nd2, compare_wcs=wcs_comp_func, tolerance=1)


def test_arithmetics_mask_func():
    def mask_sad_func(mask1, mask2, fun=0):
        if fun > 0.5:
            return mask2
        else:
            return mask1

    meta1 = {"a": 1}
    meta2 = {"a": 3, "b": 2}
    mask1 = [True, False, True]
    mask2 = [True, False, False]
    uncertainty1 = StdDevUncertainty([1, 2, 3])
    uncertainty2 = StdDevUncertainty([1, 2, 3])
    data1 = [1, 1, 1]
    data2 = [1, 1, 1]

    nd1 = NDDataArithmetic(data1, meta=meta1, mask=mask1, uncertainty=uncertainty1)
    nd2 = NDDataArithmetic(data2, meta=meta2, mask=mask2, uncertainty=uncertainty2)

    nd3 = nd1.add(nd2, handle_mask=mask_sad_func)
    assert_array_equal(nd3.mask, nd1.mask)

    nd4 = nd1.add(nd2, handle_mask=mask_sad_func, mask_fun=1)
    assert_array_equal(nd4.mask, nd2.mask)

    with pytest.raises(KeyError):
        nd1.add(nd2, handle_mask=mask_sad_func, fun=1)


@pytest.mark.parametrize("meth", ["add", "subtract", "divide", "multiply"])
def test_two_argument_useage(meth):
    ndd1 = NDDataArithmetic(np.ones((3, 3)))
    ndd2 = NDDataArithmetic(np.ones((3, 3)))

    # Call add on the class (not the instance) and compare it with already
    # tested usage:
    ndd3 = getattr(NDDataArithmetic, meth)(ndd1, ndd2)
    ndd4 = getattr(ndd1, meth)(ndd2)
    np.testing.assert_array_equal(ndd3.data, ndd4.data)

    # And the same done on an unrelated instance...
    ndd3 = getattr(NDDataArithmetic(-100), meth)(ndd1, ndd2)
    ndd4 = getattr(ndd1, meth)(ndd2)
    np.testing.assert_array_equal(ndd3.data, ndd4.data)


@pytest.mark.parametrize("meth", ["add", "subtract", "divide", "multiply"])
def test_two_argument_useage_non_nddata_first_arg(meth):
    data1 = 50
    data2 = 100

    # Call add on the class (not the instance)
    ndd3 = getattr(NDDataArithmetic, meth)(data1, data2)

    # Compare it with the instance-useage and two identical NDData-like
    # classes:
    ndd1 = NDDataArithmetic(data1)
    ndd2 = NDDataArithmetic(data2)
    ndd4 = getattr(ndd1, meth)(ndd2)
    np.testing.assert_array_equal(ndd3.data, ndd4.data)

    # and check it's also working when called on an instance
    ndd3 = getattr(NDDataArithmetic(-100), meth)(data1, data2)
    ndd4 = getattr(ndd1, meth)(ndd2)
    np.testing.assert_array_equal(ndd3.data, ndd4.data)


def test_arithmetics_unknown_uncertainties():
    # Not giving any uncertainty class means it is saved as UnknownUncertainty
    ndd1 = NDDataArithmetic(
        np.ones((3, 3)), uncertainty=UnknownUncertainty(np.ones((3, 3)))
    )
    ndd2 = NDDataArithmetic(
        np.ones((3, 3)), uncertainty=UnknownUncertainty(np.ones((3, 3)) * 2)
    )
    # There is no way to propagate uncertainties:
    with pytest.raises(IncompatibleUncertaintiesException):
        ndd1.add(ndd2)
    # But it should be possible without propagation
    ndd3 = ndd1.add(ndd2, propagate_uncertainties=False)
    np.testing.assert_array_equal(ndd1.uncertainty.array, ndd3.uncertainty.array)

    ndd4 = ndd1.add(ndd2, propagate_uncertainties=None)
    assert ndd4.uncertainty is None


def test_psf_warning():
    """Test that math on objects with a psf warn."""
    ndd1 = NDDataArithmetic(np.ones((3, 3)), psf=np.zeros(3))
    ndd2 = NDDataArithmetic(np.ones((3, 3)), psf=None)

    # no warning if both are None
    ndd2.add(ndd2)

    with pytest.warns(AstropyUserWarning, match="Not setting psf attribute during add"):
        ndd1.add(ndd2)
    with pytest.warns(AstropyUserWarning, match="Not setting psf attribute during add"):
        ndd2.add(ndd1)
    with pytest.warns(AstropyUserWarning, match="Not setting psf attribute during add"):
        ndd1.add(ndd1)


def test_raise_method_not_supported():
    ndd1 = NDDataArithmetic(np.zeros(3), uncertainty=StdDevUncertainty(np.zeros(3)))
    ndd2 = NDDataArithmetic(np.ones(3), uncertainty=StdDevUncertainty(np.ones(3)))
    result = np.zeros(3)
    correlation = 0
    # no error should be raised for supported operations:
    ndd1.uncertainty.propagate(np.add, ndd2, result, correlation)

    # raise error for unsupported propagation operations:
    with pytest.raises(ValueError):
        ndd1.uncertainty.propagate(np.mod, ndd2, result, correlation)


def test_nddata_bitmask_arithmetic():
    # NDData.mask is usually assumed to be boolean, but could be
    # a bitmask. Ensure bitmask works:
    array = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    mask = np.array([[0, 1, 64], [8, 0, 1], [2, 1, 0]])

    nref_nomask = NDDataRef(array)
    nref_masked = NDDataRef(array, mask=mask)

    # multiply no mask by constant (no mask * no mask)
    assert nref_nomask.multiply(1.0, handle_mask=np.bitwise_or).mask is None

    # multiply no mask by itself (no mask * no mask)
    assert nref_nomask.multiply(nref_nomask, handle_mask=np.bitwise_or).mask is None

    # multiply masked by constant (mask * no mask)
    np.testing.assert_equal(
        nref_masked.multiply(1.0, handle_mask=np.bitwise_or).mask, mask
    )

    # multiply masked by itself (mask * mask)
    np.testing.assert_equal(
        nref_masked.multiply(nref_masked, handle_mask=np.bitwise_or).mask, mask
    )

    # multiply masked by no mask (mask * no mask)
    np.testing.assert_equal(
        nref_masked.multiply(nref_nomask, handle_mask=np.bitwise_or).mask, mask
    )

    # check bitwise logic still works
    other_mask = np.array([[64, 1, 0], [2, 1, 0], [8, 0, 2]])
    nref_mask_other = NDDataRef(array, mask=other_mask)
    np.testing.assert_equal(
        nref_mask_other.multiply(nref_masked, handle_mask=np.bitwise_or).mask,
        np.bitwise_or(mask, other_mask),
    )


# Covers different dtypes with various types of scalars as the 2nd operand
# (issue #18384):
@pytest.mark.parametrize(
    "ndd_type",
    (
        pytest.param(np.uint16),
        pytest.param(np.float32),
        pytest.param(np.float64),
    ),
)
@pytest.mark.parametrize(
    "scalar_type",
    (
        pytest.param(int, id="int"),
        pytest.param(float, id="float"),
        pytest.param(np.uint8, id="uint8"),
        pytest.param(np.int16, id="int16"),
        pytest.param(np.float32, id="float32"),
        pytest.param(np.float64, id="float64"),
        pytest.param(lambda v: np.array(v, dtype=np.int16), id="int16_0D_array"),
        pytest.param(lambda v: np.array(v, dtype=np.float32), id="float32_0D_array"),
    ),
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_dtypes_with_scalar(ndd_type, scalar_type, meth):
    nddata = NDDataRef(np.array([1, 2, 3, 4], dtype=ndd_type))
    scalar = scalar_type(2)

    out = getattr(nddata, meth)(scalar)
    ref = STR_TO_OPERATOR[meth](nddata.data, scalar)

    # Enforce the same behaviour as NumPy, rather than fixed behaviour:
    assert out.data.shape == ref.shape
    assert out.data.dtype == ref.dtype
    assert_array_equal(out.data, ref)


# Covers adding scalar quantity with matching, non-default dtypes:
@pytest.mark.parametrize("ndd_type", (np.uint16, np.float32, np.float64))
@pytest.mark.parametrize("meth", ("add", "subtract"))
def test_add_quantity_matching_dtype(ndd_type, meth):
    nddata = NDDataRef(np.array([1, 2, 3, 4], dtype=ndd_type), unit=u.adu)
    quantity = u.Quantity(2, dtype=ndd_type, unit=u.adu)

    out = getattr(nddata, meth)(quantity)
    ref = STR_TO_OPERATOR[meth](nddata.data, quantity.value)

    assert out.data.shape == nddata.data.shape == ref.shape
    assert out.data.dtype == nddata.data.dtype == ref.dtype
    assert_array_equal(out.data, ref)


# Covers scaling with units and non-default dtypes:
@pytest.mark.parametrize("ndd_type", (np.uint16, np.float32, np.float64))
@pytest.mark.parametrize("scalar_type", (int, float, np.uint16, np.float32, np.float64))
@pytest.mark.parametrize("meth", ("multiply", "divide"))
def test_scale_dtypes_with_units(ndd_type, scalar_type, meth):
    nddata = NDDataRef(np.array([1, 2, 3, 4], dtype=ndd_type), unit=u.adu)
    scalar = scalar_type(2)

    out = getattr(nddata, meth)(scalar)
    ref = STR_TO_OPERATOR[meth](nddata.data, scalar)

    assert out.data.shape == ref.shape
    assert out.data.dtype == ref.dtype
    assert_array_almost_equal(out.data, ref)


# Covers adding scalar Quantity to NDData with default float dtypes for both.
# While these two classes live in different modules with their own defaults,
# we'd probably like to know about the unlikely event of their becoming
# inconsistent, which could break downstream assumptions. This also checks
# Quantity constructed in the common way, rather than programmatically as above.
@pytest.mark.parametrize("meth", ("add", "subtract"))
def test_add_quantity_default_dtypes(meth):
    nddata = NDDataRef(np.array([1.0, 2.0, 3.0, 4.0]), unit=u.adu)
    quantity = 2.0 * u.adu

    out = getattr(nddata, meth)(quantity)
    ref = STR_TO_OPERATOR[meth](nddata.data, quantity.value)

    assert out.data.shape == nddata.data.shape
    assert out.data.dtype == nddata.data.dtype
    assert_array_equal(out.data, ref)


# Provide input for the following test sets without lots of replication:
NDDATA_REF_PARAMS = [
    pytest.param(
        NDDataRef(
            np.array([1, 2, 3, 4], dtype="uint16"),
            uncertainty=VarianceUncertainty(np.array([1, 2, 3, 4], dtype="uint16")),
            mask=np.array([0, 1, 0, 0], dtype="uint8"),
        ),
        id="u16dat_u16var_u8msk",
    ),
    pytest.param(
        NDDataRef(
            np.array([1.0, 2.0, 3.0, 4.0], dtype="float32"),
            uncertainty=StdDevUncertainty(
                np.array([1.0, 1.41, 1.73, 2.0], dtype="float32")
            ),
            mask=np.array([0, 0, 1, 0], dtype="uint16"),
        ),
        id="f32dat_f32var_u16msk",
    ),
    pytest.param(
        NDDataRef(
            np.array([1.0, 2.0, 3.0, 4.0], dtype="float64"),
            uncertainty=VarianceUncertainty(
                np.array([1.0, 2.0, 3.0, 4.0], dtype="float64")
            ),
            mask=np.array([0, 1, 0, 0], dtype="uint16"),
        ),
        id="f64dat_f64var_u16msk",
    ),
]


@pytest.fixture(params=NDDATA_REF_PARAMS)
def nddata_ref1(request):
    return deepcopy(request.param)


@pytest.fixture(params=NDDATA_REF_PARAMS)
def nddata_ref2(request):
    return deepcopy(request.param)


# Covers non-default dtypes + uncert + mask with various scalar types
@pytest.mark.parametrize(
    "scalar_type",
    (
        pytest.param(int, id="int"),
        pytest.param(float, id="float"),
        pytest.param(np.uint16, id="uint16"),
        pytest.param(np.float32, id="float32"),
        pytest.param(lambda v: np.array(v, dtype=np.uint16), id="uint16_0D_array"),
        pytest.param(lambda v: np.array(v, dtype=np.float32), id="float32_0D_array"),
    ),
)
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_dtypes_uncert_mask_with_scalars(nddata_ref1, scalar_type, meth):
    nddata = nddata_ref1
    scalar = scalar_type(2)

    out = getattr(nddata, meth)(scalar)

    ref_dat = STR_TO_OPERATOR[meth](nddata.data, scalar)

    if meth in ("multiply", "divide"):
        vscale = scalar
        if isinstance(nddata.uncertainty, VarianceUncertainty):
            vscale = vscale * scalar  # copy to avoid modifying scalar
        ref_unc = STR_TO_OPERATOR[meth](nddata.uncertainty.array, vscale)
    else:
        ref_unc = nddata.uncertainty.array

    ref_msk = nddata.mask

    # Enforce the same behaviour as NumPy, rather than fixed behaviour:
    assert out.data.shape == ref_dat.shape
    assert out.data.dtype == ref_dat.dtype
    if not (
        NUMPY_LT_2_0
        and nddata.uncertainty.array.dtype.kind in "biu"
        and isinstance(scalar, (np.float16, np.float32))
    ):
        # A quirk of NumPy 1 arithmetic causes int uncertainty (admittedly a corner
        # case) to get cast to float64 when float32 is expected (see #18392):
        assert out.uncertainty.array.dtype == ref_unc.dtype
    assert out.mask.dtype == ref_msk.dtype
    assert np.ma.allclose(out.data, ref_dat)
    assert np.ma.allclose(out.uncertainty.array, ref_unc)
    assert_array_equal(out.mask, ref_msk)


# Covers arithmetic with different dtype pairs + uncert + mask
@pytest.mark.parametrize("meth", STR_TO_OPERATOR)
def test_arithmetics_dtypes_uncert_mask(nddata_ref1, nddata_ref2, meth):
    nd1 = nddata_ref1
    nd2 = nddata_ref2
    ref_dat = STR_TO_OPERATOR[meth](nd1.data, nd2.data)

    # Deal with uncertainty, converting the nd2 uncertainty class to match
    # nd1, otherwise arithmetic fails. With both operands being arrays, we
    # cannot use NumPy as a reference for the "correct" output dtype for
    # uncertainty, since it doesn't natively propagate errors and the result
    # type depends on the exact calculation used, but we can check that the
    # values are close those expected, given the input dtypes. Establishing
    # the intended casting behaviour for uncertainty is left for other tests.
    if isinstance(nd1.uncertainty, VarianceUncertainty):
        if isinstance(nd2.uncertainty, StdDevUncertainty):
            nd2.uncertainty = VarianceUncertainty(
                np.multiply(nd2.uncertainty.array, nd2.uncertainty.array)
            )
        if meth in ("multiply", "divide"):
            ref_unc = ref_dat**2 * (
                nd1.uncertainty.array / nd1.data**2
                + nd2.uncertainty.array / nd2.data**2
            )
        else:
            ref_unc = nd1.uncertainty.array + nd2.uncertainty.array
    else:
        if isinstance(nd2.uncertainty, VarianceUncertainty):
            nd2.uncertainty = StdDevUncertainty(np.sqrt(nd2.uncertainty.array))
        if meth in ("multiply", "divide"):
            ref_unc = ref_dat * np.sqrt(
                (nd1.uncertainty.array / nd1.data) ** 2
                + (nd2.uncertainty.array / nd2.data) ** 2
            )
        else:
            ref_unc = np.sqrt(nd1.uncertainty.array**2 + nd2.uncertainty.array**2)

    ref_msk = np.logical_or(nd1.mask, nd2.mask)  # default op for arith mixin

    out = getattr(nd1, meth)(nd2)

    # Enforce the same behaviour as NumPy, rather than fixed behaviour:
    assert out.data.shape == ref_dat.shape
    assert out.data.dtype == ref_dat.dtype
    # see above comment regarding uncertainty dtype
    assert out.mask.dtype == ref_msk.dtype
    assert np.ma.allclose(out.data, ref_dat)
    assert np.ma.allclose(out.uncertainty.array, ref_unc)
    assert_array_equal(out.mask, ref_msk)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Metadata merging."""

import warnings
from contextlib import contextmanager
from copy import deepcopy

import numpy as np

from .exceptions import MergeConflictError, MergeConflictWarning
from .utils import common_dtype

__all__ = [
    "MERGE_STRATEGIES",
    "MergeNpConcatenate",
    "MergePlus",
    "MergeStrategy",
    "enable_merge_strategies",
    "merge",
]

MERGE_STRATEGIES = []


class MergeStrategy:
    """
    Base class for defining a strategy for merging metadata from two
    sources, left and right, into a single output.

    The primary functionality for the class is the ``merge(cls, left, right)``
    class method.  This takes ``left`` and ``right`` side arguments and
    returns a single merged output.

    The first class attribute is ``types``.  This is defined as a list of
    (left_types, right_types) tuples that indicate for which input types the
    merge strategy applies.  In determining whether to apply this merge
    strategy to a pair of (left, right) objects, a test is done:
    ``isinstance(left, left_types) and isinstance(right, right_types)``.  For
    example::

      types = [(np.ndarray, np.ndarray),  # Two ndarrays
               (np.ndarray, (list, tuple)),  # ndarray and (list or tuple)
               ((list, tuple), np.ndarray)]  # (list or tuple) and ndarray

    As a convenience, ``types`` can be defined as a single two-tuple instead of
    a list of two-tuples, e.g. ``types = (np.ndarray, np.ndarray)``.

    The other class attribute is ``enabled``, which defaults to ``False`` in
    the base class.  By defining a subclass of ``MergeStrategy`` the new merge
    strategy is automatically registered to be available for use in
    merging. However, by default the new merge strategy is *not enabled*.  This
    prevents inadvertently changing the behavior of unrelated code that is
    performing metadata merge operations.

    In most cases (particularly in library code that others might use) it is
    recommended to leave custom strategies disabled and use the
    `~astropy.utils.metadata.enable_merge_strategies` context manager to locally
    enable the desired strategies.  However, if one is confident that the
    new strategy will not produce unexpected behavior, then one can globally
    enable it by setting the ``enabled`` class attribute to ``True``.

    Examples
    --------
    Here we define a custom merge strategy that takes an int or float on
    the left and right sides and returns a list with the two values.

      >>> from astropy.utils.metadata import MergeStrategy
      >>> class MergeNumbersAsList(MergeStrategy):
      ...     types = ((int, float), (int, float))  # (left_types, right_types)
      ...
      ...     @classmethod
      ...     def merge(cls, left, right):
      ...         return [left, right]

    """

    # Set ``enabled = True`` to globally enable applying this merge strategy.
    # This is not generally recommended.
    enabled = False

    # types = [(left_types, right_types), ...]

    def __init_subclass__(cls):
        members = vars(cls)
        # Register merging class (except for base MergeStrategy class)
        if (types := members.get("types")) is not None:
            if isinstance(types, tuple):
                types = [types]
            for left, right in reversed(types):
                MERGE_STRATEGIES.insert(0, (left, right, cls))

    @classmethod
    def _merge(cls, left, right):
        try:
            return cls.merge(left, right)
        except Exception as err:
            raise MergeConflictError(err)


class MergePlus(MergeStrategy):
    """
    Merge ``left`` and ``right`` objects using the plus operator.  This
    merge strategy is globally enabled by default.
    """

    types = [(list, list), (tuple, tuple)]
    enabled = True

    @classmethod
    def merge(cls, left, right):
        return left + right


class MergeNpConcatenate(MergeStrategy):
    """
    Merge ``left`` and ``right`` objects using np.concatenate.  This
    merge strategy is globally enabled by default.

    This will upcast a list or tuple to np.ndarray and the output is
    always ndarray.
    """

    types = [
        (np.ndarray, np.ndarray),
        (np.ndarray, (list, tuple)),
        ((list, tuple), np.ndarray),
    ]
    enabled = True

    @classmethod
    def merge(cls, left, right):
        left, right = np.asanyarray(left), np.asanyarray(right)
        common_dtype([left, right])  # Ensure left and right have compatible dtype
        return np.concatenate([left, right])


# ============================================================================


@contextmanager
def enable_merge_strategies(*merge_strategies):
    """
    Context manager to temporarily enable one or more custom metadata merge
    strategies.

    Examples
    --------
    Here we define a custom merge strategy that takes an int or float on
    the left and right sides and returns a list with the two values.

      >>> from astropy.utils.metadata import MergeStrategy
      >>> class MergeNumbersAsList(MergeStrategy):
      ...     types = ((int, float),  # left side types
      ...              (int, float))  # right side types
      ...     @classmethod
      ...     def merge(cls, left, right):
      ...         return [left, right]

    By defining this class the merge strategy is automatically registered to be
    available for use in merging. However, by default new merge strategies are
    *not enabled*.  This prevents inadvertently changing the behavior of
    unrelated code that is performing metadata merge operations.

    In order to use the new merge strategy, use this context manager as in the
    following example::

      >>> from astropy.table import Table, vstack
      >>> from astropy.utils.metadata import enable_merge_strategies
      >>> t1 = Table([[1]], names=['a'])
      >>> t2 = Table([[2]], names=['a'])
      >>> t1.meta = {'m': 1}
      >>> t2.meta = {'m': 2}
      >>> with enable_merge_strategies(MergeNumbersAsList):
      ...    t12 = vstack([t1, t2])
      >>> t12.meta['m']
      [1, 2]

    One can supply further merge strategies as additional arguments to the
    context manager.

    As a convenience, the enabling operation is actually done by checking
    whether the registered strategies are subclasses of the context manager
    arguments.  This means one can define a related set of merge strategies and
    then enable them all at once by enabling the base class.  As a trivial
    example, *all* registered merge strategies can be enabled with::

      >>> with enable_merge_strategies(MergeStrategy):
      ...    t12 = vstack([t1, t2])

    Parameters
    ----------
    *merge_strategies : :class:`~astropy.utils.metadata.MergeStrategy` class
        Merge strategies that will be enabled.
    """
    orig_enabled = {}
    for _, _, merge_strategy in MERGE_STRATEGIES:
        if issubclass(merge_strategy, merge_strategies):
            orig_enabled[merge_strategy] = merge_strategy.enabled
            merge_strategy.enabled = True

    yield

    for merge_strategy, enabled in orig_enabled.items():
        merge_strategy.enabled = enabled


# =============================================================================


def _both_isinstance(left, right, cls):
    return isinstance(left, cls) and isinstance(right, cls)


def _not_equal(left, right):
    try:
        return bool(left != right)
    except Exception:
        return True


def _warn_str_func(key, left, right):
    out = (
        f"Cannot merge meta key {key!r} types {type(left)!r}"
        f" and {type(right)!r}, choosing {key}={right!r}"
    )
    return out


def _error_str_func(key, left, right):
    out = f"Cannot merge meta key {key!r} types {type(left)!r} and {type(right)!r}"
    return out


def _raise_merge_conflict_error(left, right):
    raise MergeConflictError


def merge(
    left,
    right,
    merge_func=None,
    metadata_conflicts="warn",
    warn_str_func=_warn_str_func,
    error_str_func=_error_str_func,
):
    """
    Merge the ``left`` and ``right`` metadata objects.

    This is a simplistic and limited implementation at this point.
    """
    if not _both_isinstance(left, right, dict):
        raise MergeConflictError("Can only merge two dict-based objects")

    out = deepcopy(left)

    for key, right_val in right.items():
        # If no conflict then insert val into out dict and continue
        if key not in out:
            out[key] = deepcopy(right_val)
            continue

        # There is a conflict that must be resolved
        left_val = out[key]
        if _both_isinstance(left_val, right_val, dict):
            out[key] = merge(
                left_val, right_val, merge_func, metadata_conflicts=metadata_conflicts
            )
            continue

        local_merge_func = merge_func
        if local_merge_func is None:
            for left_type, right_type, merge_cls in MERGE_STRATEGIES:
                if (
                    merge_cls.enabled
                    and isinstance(left_val, left_type)
                    and isinstance(right_val, right_type)
                ):
                    local_merge_func = merge_cls._merge
                    break
            else:
                local_merge_func = _raise_merge_conflict_error
        try:
            out[key] = local_merge_func(left_val, right_val)
        except MergeConflictError:
            # If right_val is None or equal to left_val then we are happy with what's
            # already in out. Otherwise right_val takes priority (or there's an error).
            if right_val is not None and _not_equal(left_val, right_val):
                if metadata_conflicts == "warn":
                    warnings.warn(
                        warn_str_func(key, left_val, right_val), MergeConflictWarning
                    )
                elif metadata_conflicts == "error":
                    raise MergeConflictError(error_str_func(key, left_val, right_val))
                elif metadata_conflicts != "silent":
                    raise ValueError(
                        "metadata_conflicts argument must be one "
                        'of "silent", "warn", or "error"'
                    )
                out[key] = right_val
    return out
